"""Unit tests for :mod:`~tmlt.core.transformations.spark_transformations.map`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from typing import Callable, Dict, List, Optional, Union

import pandas as pd
import sympy as sp
from parameterized import parameterized
from pyspark import Row

from tmlt.core.domains.base import Domain
from tmlt.core.domains.collections import ListDomain
from tmlt.core.domains.spark_domains import (
    SparkColumnDescriptor,
    SparkDataFrameDomain,
    SparkFloatColumnDescriptor,
    SparkIntegerColumnDescriptor,
    SparkRowDomain,
    SparkStringColumnDescriptor,
)
from tmlt.core.metrics import (
    HammingDistance,
    IfGroupedBy,
    NullMetric,
    RootSumOfSquared,
    SumOf,
    SymmetricDifference,
)
from tmlt.core.transformations.spark_transformations.map import (
    FlatMap,
    GroupingFlatMap,
    Map,
    RowToRowsTransformation,
    RowToRowTransformation,
)
from tmlt.core.utils.testing import (
    TestComponent,
    assert_property_immutability,
    get_all_props,
)


class TestRowToRowsTransformer(TestComponent):
    """Tests for class RowToRowsTransformation.

    Tests
    :class:`~tmlt.core.transformations.spark_transformations.map.
    RowToRowsTransformation`.
    """

    @parameterized.expand(get_all_props(RowToRowsTransformation))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        transformer = RowToRowsTransformation(
            SparkRowDomain(self.schema_a),
            ListDomain(SparkRowDomain(self.schema_a)),
            lambda x: [x],
            False,
        )
        assert_property_immutability(transformer, prop_name)

    @parameterized.expand([(True,), (False,)])
    def test_properties(self, augment: bool):
        """RowToRowsTransformation's properties have the expected values."""
        input_domain = SparkRowDomain(self.schema_a)
        transformer = RowToRowsTransformation(
            input_domain, ListDomain(input_domain), lambda x: [x], augment
        )
        self.assertEqual(transformer.input_domain, input_domain)
        self.assertEqual(transformer.input_metric, NullMetric())
        self.assertEqual(transformer.output_domain, ListDomain(input_domain))
        self.assertEqual(transformer.output_metric, NullMetric())
        self.assertEqual(transformer.trusted_f(Row(x=5)), [Row(x=5)])
        self.assertEqual(transformer.augment, augment)

    @parameterized.expand(
        [
            (
                lambda x: [x, x],
                False,
                [Row(A=4.2, B="X"), Row(A=4.2, B="X")],
                ListDomain(
                    SparkRowDomain(
                        {
                            "A": SparkFloatColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                        }
                    )
                ),
            ),
            (
                lambda x: [{"A2": 2 * x["A"]}],
                False,
                [Row(A2=8.4)],
                ListDomain(SparkRowDomain({"A2": SparkFloatColumnDescriptor()})),
            ),
            (
                lambda x: [{"A2": 2 * x["A"]}],
                True,
                [Row(A=4.2, B="X", A2=8.4)],
                ListDomain(
                    SparkRowDomain(
                        {
                            "A": SparkFloatColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                            "A2": SparkStringColumnDescriptor(),
                        }
                    )
                ),
            ),
        ]
    )
    def test_transformer_works_correctly(
        self, f: Callable, augment: bool, expected: List, output_domain: ListDomain
    ):
        """Tests that row transformer works correctly."""
        transformer = RowToRowsTransformation(
            SparkRowDomain(self.schema_a), output_domain, f, augment
        )

        self.assertEqual(transformer.input_domain, SparkRowDomain(self.schema_a))
        self.assertEqual(transformer.output_domain, output_domain)
        row = Row(A=4.2, B="X")
        result_rows = transformer(row)
        self.assertEqual(result_rows, expected)


class TestMap(TestComponent):
    """Tests for :class:`~tmlt.core.transformations.spark_transformations.map.Map`."""

    @parameterized.expand(get_all_props(Map))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        t = Map(
            metric=SymmetricDifference(),
            row_transformer=RowToRowTransformation(
                input_domain=SparkRowDomain(self.schema_a),
                output_domain=SparkRowDomain(self.schema_a),
                trusted_f=lambda row: row,
                augment=False,
            ),
        )
        assert_property_immutability(t, prop_name)

    def test_properties(self):
        """Map's properties have the expected values."""
        times_two = lambda row: {"A": row["A"] * 2, "B": row["B"]}
        row_transformer = RowToRowTransformation(
            input_domain=SparkRowDomain(self.schema_a),
            output_domain=SparkRowDomain(self.schema_a),
            trusted_f=times_two,
            augment=False,
        )
        transformation = Map(
            metric=SymmetricDifference(), row_transformer=row_transformer
        )
        self.assertEqual(
            transformation.input_domain, SparkDataFrameDomain(self.schema_a)
        )
        self.assertEqual(transformation.input_metric, SymmetricDifference())
        self.assertEqual(
            transformation.output_domain, SparkDataFrameDomain(self.schema_a)
        )
        self.assertEqual(transformation.output_metric, SymmetricDifference())
        self.assertEqual(transformation.row_transformer, row_transformer)

    def test_map(self):
        """Tests that map transformation works correctly."""
        times_two = lambda row: {"A": row["A"] * 2, "B": row["B"]}
        times_two_transformation = Map(
            metric=SymmetricDifference(),
            row_transformer=RowToRowTransformation(
                input_domain=SparkRowDomain(self.schema_a),
                output_domain=SparkRowDomain(self.schema_a),
                trusted_f=times_two,
                augment=False,
            ),
        )
        self.assertEqual(times_two_transformation.stability_function(1), 1)
        self.assertTrue(times_two_transformation.stability_relation(1, 1))
        actual_df = times_two_transformation(self.df_a).toPandas()
        expected_df = pd.DataFrame([[2.4, "X"]], columns=["A", "B"])
        self.assert_frame_equal_with_sort(actual_df, expected_df)

    def test_map_empty(self):
        """Tests that map transformation works correctly."""
        times_two = lambda row: {"A": row["A"] * 2, "B": row["B"]}
        times_two_transformation = Map(
            metric=SymmetricDifference(),
            row_transformer=RowToRowTransformation(
                input_domain=SparkRowDomain(self.schema_a),
                output_domain=SparkRowDomain(self.schema_a),
                trusted_f=times_two,
                augment=False,
            ),
        )
        empty_df = pd.DataFrame([], columns=["A", "B"])
        actual_df = times_two_transformation(
            self.spark.createDataFrame(empty_df, self.df_a.schema)
        ).toPandas()
        expected_df = empty_df
        self.assert_frame_equal_with_sort(actual_df, expected_df)

    @parameterized.expand(
        [
            (SymmetricDifference(),),
            (HammingDistance(),),
            (IfGroupedBy("B", SumOf(SymmetricDifference())),),
            (IfGroupedBy("B", RootSumOfSquared(SymmetricDifference())),),
            (IfGroupedBy("B", SymmetricDifference()),),
        ]
    )
    def test_metrics(self, metric: Union[SymmetricDifference, IfGroupedBy]):
        """Tests that Map works correctly with supported metrics."""
        id_map = Map(
            metric=metric,
            row_transformer=RowToRowTransformation(
                input_domain=SparkRowDomain(self.schema_a),
                output_domain=SparkRowDomain(self.schema_a),
                trusted_f=lambda row: row,
                augment=True,
            ),
        )
        self.assertTrue(id_map.input_metric == metric == id_map.output_metric)
        self.assertTrue(id_map.stability_function(1), 1)
        self.assertTrue(id_map.stability_relation(1, 1))
        actual = id_map(self.df_a).toPandas()
        self.assert_frame_equal_with_sort(actual, self.df_a.toPandas())

    @parameterized.expand(
        [
            (
                "C",
                RootSumOfSquared(SymmetricDifference()),
                True,
                "Input metric .* and input domain .* are not compatible",
            ),  # Invalid column
            (
                "B",
                RootSumOfSquared(SymmetricDifference()),
                False,
                "Transformer must be augmenting",
            ),  # Does not augment
            (
                "B",
                SumOf(HammingDistance()),
                True,
                "must be SymmetricDifference",
            ),  # Unsupported inner metric
        ]
    )
    def test_if_grouped_by_metric_invalid_parameters(
        self,
        groupby_column: str,
        inner_metric: Union[SumOf, RootSumOfSquared, SymmetricDifference],
        augment: bool,
        error_msg: str,
    ):
        """Tests that Map raises appropriate error with invalid parameters."""
        with self.assertRaisesRegex(ValueError, error_msg):
            Map(
                metric=IfGroupedBy(groupby_column, inner_metric),
                row_transformer=RowToRowTransformation(
                    input_domain=SparkRowDomain(self.schema_a),
                    output_domain=SparkRowDomain(self.schema_a),
                    trusted_f=lambda row: row,
                    augment=augment,
                ),
            )


class TestFlatMap(TestComponent):
    """Tests for class FlatMap.

    Tests :class:`~tmlt.core.transformations.spark_transformations.map.FlatMap`.
    """

    @parameterized.expand(get_all_props(FlatMap))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        t = FlatMap(
            metric=SymmetricDifference(),
            row_transformer=self.duplicate_transformer,
            max_num_rows=1,
        )
        assert_property_immutability(t, prop_name)

    def test_properties(self):
        """FlatMap's properties have the expected values."""
        duplicate = lambda row: [row, row]
        row_transformer = RowToRowsTransformation(
            input_domain=SparkRowDomain(self.schema_a),
            output_domain=ListDomain(SparkRowDomain(self.schema_a)),
            trusted_f=duplicate,
            augment=False,
        )
        transformation = FlatMap(
            metric=SymmetricDifference(),
            row_transformer=row_transformer,
            max_num_rows=2,
        )
        self.assertEqual(
            transformation.input_domain, SparkDataFrameDomain(self.schema_a)
        )
        self.assertEqual(transformation.input_metric, SymmetricDifference())
        self.assertEqual(
            transformation.output_domain, SparkDataFrameDomain(self.schema_a)
        )
        self.assertEqual(transformation.output_metric, SymmetricDifference())
        self.assertEqual(transformation.row_transformer, row_transformer)
        self.assertEqual(transformation.max_num_rows, 2)

    @parameterized.expand(
        [
            (1, pd.DataFrame([[1.2, "X"]], columns=["A", "B"])),
            (2, pd.DataFrame([[1.2, "X"], [1.2, "X"]], columns=["A", "B"])),
            (3, pd.DataFrame([[1.2, "X"], [1.2, "X"]], columns=["A", "B"])),
            (0, pd.DataFrame([], columns=["A", "B"])),
        ]
    )
    def test_flat_map(self, max_num_rows: int, expected_df: pd.DataFrame):
        """Tests that flat_map transformation works correctly."""
        flat_map_transformation = FlatMap(
            metric=SymmetricDifference(),
            row_transformer=self.duplicate_transformer,
            max_num_rows=max_num_rows,
        )
        self.assertEqual(flat_map_transformation.stability_function(1), max_num_rows)
        self.assertTrue(flat_map_transformation.stability_relation(1, max_num_rows))
        actual_df = flat_map_transformation(self.df_a).toPandas()
        self.assert_frame_equal_with_sort(actual_df, expected_df)

    @parameterized.expand(
        [
            (
                IfGroupedBy("B", SumOf(SymmetricDifference())),
                1,
                pd.DataFrame([[1.2, "X"]], columns=["A", "B"]),
                1,
            ),
            (
                IfGroupedBy("B", SumOf(SymmetricDifference())),
                2,
                pd.DataFrame([[1.2, "X"], [1.2, "X"]], columns=["A", "B"]),
                2,
            ),
            (
                IfGroupedBy("B", SymmetricDifference()),
                0,
                pd.DataFrame([], columns=["A", "B"]),
                1,
            ),
            (
                IfGroupedBy("B", SymmetricDifference()),
                None,
                pd.DataFrame([[1.2, "X"], [1.2, "X"]], columns=["A", "B"]),
                1,
            ),
        ]
    )
    def test_flat_map_augmenting(
        self,
        metric: IfGroupedBy,
        max_num_rows: int,
        expected_df: pd.DataFrame,
        stability: int,
    ):
        """Tests that flat_map transformation works correctly on IfGroupedBy metrics."""

        flat_map_transformation = FlatMap(
            metric=metric,
            row_transformer=self.augmenting_duplicate_transformer,
            max_num_rows=max_num_rows,
        )
        self.assertEqual(flat_map_transformation.stability_function(1), stability)
        self.assertTrue(flat_map_transformation.stability_relation(1, stability))
        actual_df = flat_map_transformation(self.df_a).toPandas()
        self.assert_frame_equal_with_sort(actual_df, expected_df)

    @parameterized.expand(
        [
            (  # Bad input_domain
                ListDomain(SparkRowDomain({})),
                ListDomain(SparkRowDomain({})),
            ),
            (  # Bad output_domain.element_domain
                SparkRowDomain({}),
                ListDomain(SparkDataFrameDomain({})),
            ),
            (SparkRowDomain({}), SparkDataFrameDomain({})),  # Bad output_domain
        ]
    )
    def test_domains(self, input_domain: Domain, output_domain: Domain):
        """Tests that the constructor checks domains correctly.

        RowToRowsTransformation must meet the following conditions
            - input_domain is a *SparkRowDomain*
            - output_domain is a *ListDomain*
            - output_domain.element_domain is a *SparkRowDomain*
        """
        with self.assertRaises((TypeError, ValueError)):
            FlatMap(
                metric=SymmetricDifference(),
                row_transformer=RowToRowsTransformation(
                    input_domain,  # type: ignore
                    output_domain,  # type: ignore
                    lambda x: [x],
                    augment=False,
                ),
                max_num_rows=1,
            )

    @parameterized.expand(
        [
            (SymmetricDifference(), 2, 2),
            (IfGroupedBy("B", SumOf(SymmetricDifference())), 2, 2),
            (IfGroupedBy("B", RootSumOfSquared(SymmetricDifference())), 2, 2),
            (IfGroupedBy("B", SymmetricDifference()), 1, 2),
            (IfGroupedBy("B", SymmetricDifference()), 1, None),
        ]
    )
    def test_metrics(
        self,
        metric: Union[SymmetricDifference, IfGroupedBy],
        d_out: int,
        max_num_rows: Optional[int],
    ):
        """Tests that FlatMap works correctly with supported metrics."""
        split_map = FlatMap(
            metric=metric,
            row_transformer=RowToRowsTransformation(
                input_domain=SparkRowDomain(self.schema_a),
                output_domain=ListDomain(
                    SparkRowDomain(
                        {**self.schema_a, "C": SparkIntegerColumnDescriptor()}
                    )
                ),
                trusted_f=lambda row: [
                    {**row.asDict(), "C": 1},
                    {**row.asDict(), "C": 2},
                ],
                augment=True,
            ),
            max_num_rows=max_num_rows,
        )
        self.assertTrue(split_map.input_metric == metric == split_map.output_metric)
        self.assertTrue(split_map.stability_function(1), d_out)
        self.assertTrue(split_map.stability_relation(1, d_out))
        actual = split_map(self.df_a).toPandas()
        expected = pd.DataFrame({"A": [1.2, 1.2], "B": ["X", "X"], "C": [1, 2]})
        self.assert_frame_equal_with_sort(actual, expected)

    @parameterized.expand(
        [
            (
                "C",
                True,
                "Input metric .* and input domain .* are not compatible",
                SymmetricDifference(),
            ),  # Invalid column
            (
                "B",
                False,
                "Transformer must be augmenting",
                SymmetricDifference(),
            ),  # Does not augment
            (
                "B",
                True,
                "must be SymmetricDifference",
                HammingDistance(),
            ),  # Does not augment
        ]
    )
    def test_if_grouped_by_metric_invalid_parameters(
        self,
        groupby_column: str,
        augment: bool,
        error_msg: str,
        inner_metric: Union[SymmetricDifference, HammingDistance],
    ):
        """Tests that Map raises appropriate error with invalid parameters."""
        with self.assertRaisesRegex(ValueError, error_msg):
            FlatMap(
                metric=IfGroupedBy(groupby_column, SumOf(inner_metric)),
                row_transformer=RowToRowsTransformation(
                    input_domain=SparkRowDomain(self.schema_a),
                    output_domain=ListDomain(SparkRowDomain(self.schema_a)),
                    trusted_f=lambda row: [row],
                    augment=augment,
                ),
                max_num_rows=2,
            )

    @parameterized.expand(
        [
            (SymmetricDifference(),),
            (IfGroupedBy("B", SumOf(SymmetricDifference())),),
            (IfGroupedBy("B", RootSumOfSquared(SymmetricDifference())),),
        ]
    )
    def test_none_stability(self, metric: Union[SymmetricDifference, IfGroupedBy]):
        """Tests that FlatMap raises exaception when None stability not allowed."""
        with self.assertRaisesRegex(
            ValueError, "max_num_rows must be specified if metric is not"
        ):
            FlatMap(
                metric=metric,
                row_transformer=RowToRowsTransformation(
                    input_domain=SparkRowDomain(self.schema_a),
                    output_domain=ListDomain(SparkRowDomain(self.schema_a)),
                    trusted_f=lambda row: [row],
                    augment=True,
                ),
                max_num_rows=None,
            )


class TestGroupingFlatMap(TestComponent):
    """Tests for class GroupingFlatMap.

    Tests
    :class:`~tmlt.core.transformations.spark_transformations.map.GroupingFlatMap`.
    """

    @parameterized.expand(get_all_props(GroupingFlatMap))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        duplicate = lambda row: [{**row.asDict(), "G": 1}, {**row.asDict(), "G": 2}]
        row_transformer = RowToRowsTransformation(
            input_domain=SparkRowDomain(self.schema_a),
            output_domain=ListDomain(
                SparkRowDomain({**self.schema_a, "G": SparkIntegerColumnDescriptor()})
            ),
            trusted_f=duplicate,
            augment=True,
        )
        t = GroupingFlatMap(
            output_metric=RootSumOfSquared(SymmetricDifference()),
            row_transformer=row_transformer,
            max_num_rows=2,
        )
        assert_property_immutability(t, prop_name)

    def test_properties(self):
        """GroupingFlatMap's properties have the expected values."""
        duplicate = lambda row: [{**row.asDict(), "G": 1}, {**row.asDict(), "G": 2}]
        output_schema: Dict[str, SparkColumnDescriptor] = {
            **self.schema_a,
            "G": SparkIntegerColumnDescriptor(),
        }
        row_transformer = RowToRowsTransformation(
            input_domain=SparkRowDomain(self.schema_a),
            output_domain=ListDomain(SparkRowDomain(output_schema)),
            trusted_f=duplicate,
            augment=True,
        )
        transformation = GroupingFlatMap(
            output_metric=RootSumOfSquared(SymmetricDifference()),
            row_transformer=row_transformer,
            max_num_rows=2,
        )
        self.assertEqual(
            transformation.input_domain, SparkDataFrameDomain(self.schema_a)
        )
        self.assertEqual(transformation.input_metric, SymmetricDifference())
        self.assertEqual(
            transformation.output_domain, SparkDataFrameDomain(output_schema)
        )
        self.assertEqual(
            transformation.output_metric,
            IfGroupedBy("G", RootSumOfSquared(SymmetricDifference())),
        )
        self.assertEqual(transformation.row_transformer, row_transformer)
        self.assertEqual(transformation.max_num_rows, 2)

    @parameterized.expand(
        [
            (
                lambda row: [{**row.asDict(), "NEW": 1}, {**row.asDict(), "NEW": 2}],
                2,
                pd.DataFrame(
                    [("A", 1), ("A", 2), ("B", 1), ("B", 2)], columns=["X", "NEW"]
                ),
            ),
            (  # Duplicate group value
                lambda row: [{**row.asDict(), "NEW": 1}, {**row.asDict(), "NEW": 1}],
                2,
                pd.DataFrame([("A", 1), ("B", 1)], columns=["X", "NEW"]),
            ),
            (  # Truncates correctly
                lambda row: [{**row.asDict(), "NEW": 1}, {**row.asDict(), "NEW": 2}],
                1,
                pd.DataFrame([("A", 1), ("B", 1)], columns=["X", "NEW"]),
            ),
            (  # Does not overwrite
                lambda row: [{"X": "C", "NEW": 1}, {"X": "C", "NEW": 2}],
                2,
                pd.DataFrame(
                    [("A", 1), ("A", 2), ("B", 1), ("B", 2)], columns=["X", "NEW"]
                ),
            ),
        ]
    )
    def test_grouping_flatmap(
        self, trusted_f: Callable, max_num_rows: int, expected_df
    ):
        """Tests that GroupingFlatMap works correctly."""
        df = self.spark.createDataFrame(pd.DataFrame({"X": ["A", "B"]}))
        transformation = GroupingFlatMap(
            output_metric=RootSumOfSquared(SymmetricDifference()),
            row_transformer=RowToRowsTransformation(
                input_domain=SparkRowDomain({"X": SparkStringColumnDescriptor()}),
                output_domain=ListDomain(
                    SparkRowDomain(
                        {
                            "X": SparkStringColumnDescriptor(),
                            "NEW": SparkIntegerColumnDescriptor(),
                        }
                    )
                ),
                trusted_f=trusted_f,
                augment=True,
            ),
            max_num_rows=max_num_rows,
        )
        self.assert_frame_equal_with_sort(transformation(df).toPandas(), expected_df)

    @parameterized.expand([(1,), (2,), (4,), (9,)])
    def test_stability_function_and_relation(self, max_num_rows: int):
        """Tests that stability relation works correctly."""
        transformation = GroupingFlatMap(
            output_metric=RootSumOfSquared(SymmetricDifference()),
            row_transformer=RowToRowsTransformation(
                input_domain=SparkRowDomain(
                    {
                        "A": SparkStringColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                ),
                output_domain=ListDomain(
                    SparkRowDomain(
                        {
                            "A": SparkStringColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                            "C": SparkStringColumnDescriptor(),
                        }
                    )
                ),
                trusted_f=lambda row: [{**row.asDict(), "C": "1"}],
                augment=True,
            ),
            max_num_rows=max_num_rows,
        )
        self.assertEqual(transformation.stability_function(1), sp.sqrt(max_num_rows))
        self.assertTrue(transformation.stability_relation(1, sp.sqrt(max_num_rows)))

    @parameterized.expand(
        [
            (
                RowToRowsTransformation(
                    input_domain=SparkRowDomain({"A": SparkIntegerColumnDescriptor()}),
                    output_domain=ListDomain(
                        SparkRowDomain(
                            {
                                "A": SparkIntegerColumnDescriptor(),
                                "B": SparkIntegerColumnDescriptor(),
                            }
                        )
                    ),
                    trusted_f=lambda row: [row],
                    augment=False,
                ),
                "Transformer must be augmenting",
            ),
            (
                RowToRowsTransformation(
                    input_domain=SparkRowDomain({"A": SparkIntegerColumnDescriptor()}),
                    output_domain=ListDomain(
                        SparkRowDomain({"A": SparkIntegerColumnDescriptor()})
                    ),
                    trusted_f=lambda row: [row],
                    augment=True,
                ),
                "No grouping column provided",
            ),
            (
                RowToRowsTransformation(
                    input_domain=SparkRowDomain({"A": SparkIntegerColumnDescriptor()}),
                    output_domain=ListDomain(
                        SparkRowDomain(
                            {
                                "A": SparkIntegerColumnDescriptor(),
                                "B": SparkIntegerColumnDescriptor(),
                                "C": SparkIntegerColumnDescriptor(),
                            }
                        )
                    ),
                    trusted_f=lambda row: [row],
                    augment=True,
                ),
                "Only one grouping column allowed",
            ),
        ]
    )
    def test_invalid_transformers_fail(
        self, transformer: RowToRowsTransformation, error_msg: str
    ):
        """Tests that invalid (nonaugmenting, no extra, more than 1 extra)."""
        with self.assertRaisesRegex(ValueError, error_msg):
            GroupingFlatMap(
                output_metric=RootSumOfSquared(SymmetricDifference()),
                row_transformer=transformer,
                max_num_rows=1,
            )



"""Base class for transformations."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from abc import ABC, abstractmethod
from typing import Any, Union, overload

from typeguard import check_type, typechecked

from tmlt.core.domains.base import Domain
from tmlt.core.measurements.base import Measurement
from tmlt.core.metrics import Metric, UnsupportedCombinationError


class Transformation(ABC):
    """Abstract base class for transformations."""

    @typechecked
    def __init__(
        self,
        input_domain: Domain,
        input_metric: Metric,
        output_domain: Domain,
        output_metric: Metric,
    ):
        """Base constructor for transformations."""
        if not input_metric.supports_domain(input_domain):
            raise UnsupportedCombinationError(
                (input_metric, input_domain),
                (
                    f"Input metric {input_metric} and input domain {input_domain} are"
                    " not compatible."
                ),
            )
        if not output_metric.supports_domain(output_domain):
            raise UnsupportedCombinationError(
                (output_metric, output_domain),
                (
                    f"Output metric {output_metric} and output domain"
                    f" {output_domain} are not compatible."
                ),
            )
        self._input_domain = input_domain
        self._input_metric = input_metric
        self._output_domain = output_domain
        self._output_metric = output_metric

    @property
    def input_domain(self) -> Domain:
        """Return input domain for the measurement."""
        return self._input_domain

    @property
    def input_metric(self) -> Metric:
        """Distance metric on input domain."""
        return self._input_metric

    @property
    def output_domain(self) -> Domain:
        """Return input domain for the measurement."""
        return self._output_domain

    @property
    def output_metric(self) -> Metric:
        """Distance metric on input domain."""
        return self._output_metric

    @typechecked
    def stability_function(self, d_in: Any) -> Any:
        """Returns the smallest d_out satisfied by the transformation.

        See the privacy and stability tutorial (add link?) for more information.

        Args:
            d_in: Distance between inputs under input_metric.

        Raises:
            NotImplementedError: If not overridden.
        """
        self.input_metric.validate(d_in)
        raise NotImplementedError(
            f"{self.__class__.__name__} does not have a stability function"
        )
        return d_in  # pylint: disable=unreachable

    @typechecked
    def stability_relation(self, d_in: Any, d_out: Any) -> bool:
        """Returns True only if close inputs produce close outputs.

        See the privacy and stability tutorial (add link?) for more information.

        Args:
            d_in: Distance between inputs under input_metric.
            d_out: Distance between outputs under output_metric.
        """
        min_d_out = self.stability_function(d_in)
        if min_d_out is NotImplemented:
            raise NotImplementedError()
        return self.output_metric.compare(min_d_out, d_out)

    @overload
    def __or__(
        self, other: "Transformation"
    ) -> "Transformation":  # noqa: D105 https://github.com/PyCQA/pydocstyle/issues/525
        ...

    @overload
    def __or__(
        self, other: Measurement
    ) -> Measurement:  # noqa: D105 https://github.com/PyCQA/pydocstyle/issues/525
        ...

    def __or__(self, other):
        """Return this transformation chained with another component."""
        # pylint: disable=import-outside-toplevel
        check_type("other", other, Union[Measurement, Transformation])
        if isinstance(other, Measurement):
            from tmlt.core.measurements.chaining import ChainTM

            return ChainTM(transformation=self, measurement=other)
        assert isinstance(other, Transformation)
        from tmlt.core.transformations.chaining import ChainTT

        return ChainTT(transformation1=self, transformation2=other)

    @abstractmethod
    def __call__(self, data: Any) -> Any:
        """Perform transformation."""



"""Tests for :mod:`~tmlt.core.util.grouped_dataframe`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

import pandas as pd
from parameterized import parameterized
from pyspark.sql import Row
from pyspark.sql import functions as sf
from pyspark.sql.types import IntegerType, StructField, StructType

from tmlt.core.utils.grouped_dataframe import GroupedDataFrame
from tmlt.core.utils.testing import (
    PySparkTest,
    assert_property_immutability,
    get_all_props,
)

# pylint: disable=no-member


class TestGroupedDataFrame(PySparkTest):
    """Tests for GroupedDataFrame."""

    @parameterized.expand(get_all_props(GroupedDataFrame))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        grouped_df = GroupedDataFrame(
            dataframe=self.spark.createDataFrame(
                [("A", 1), ("B", 2)], schema=["X", "Y"]
            ),
            group_keys=self.spark.createDataFrame([("A",), ("B",)], schema=["X"]),
        )
        assert_property_immutability(grouped_df, prop_name)

    @parameterized.expand(
        [
            (
                pd.DataFrame([(1, 2)], columns=["A", "Z"]),
                pd.DataFrame([(1,)], columns=["B"]),
                "Invalid groupby columns",
            )
        ]
    )
    def test_constructor_invalid_inputs(
        self, dataframe: pd.DataFrame, group_keys: pd.DataFrame, error_msg: str
    ):
        """Tests that error is raised when constructor called with invalid inputs."""
        with self.assertRaisesRegex(ValueError, error_msg):
            GroupedDataFrame(
                dataframe=self.spark.createDataFrame(dataframe),
                group_keys=self.spark.createDataFrame(group_keys),
            )

    def test_constructor_drops_duplicate_group_keys(self):
        """Tests that duplicate group keys are silently dropped."""
        grouped_dataframe = GroupedDataFrame(
            dataframe=self.spark.createDataFrame([(1, 2)], schema=["A", "B"]),
            group_keys=self.spark.createDataFrame([(1,), (1,)], schema=["A"]),
        )
        expected_group_keys = pd.DataFrame({"A": [1]})
        self.assert_frame_equal_with_sort(
            expected_group_keys, grouped_dataframe.group_keys.toPandas()
        )

    def test_select_works_correctly(self):
        """Tests that select works correctly."""
        grouped_dataframe = GroupedDataFrame(
            dataframe=self.spark.createDataFrame([(1, 2, 3)], schema=["A", "B", "C"]),
            group_keys=self.spark.createDataFrame([(1,), (1,)], schema=["A"]),
        )
        expected = pd.DataFrame({"A": [1], "B": [2]})
        actual = grouped_dataframe.select(  # pylint:disable=protected-access
            ["A", "B"]
        )._dataframe.toPandas()
        self.assert_frame_equal_with_sort(actual, expected)

    def test_agg_with_nulls(self) -> None:
        """Test that .agg works correctly with nulls."""
        data = pd.DataFrame({"A": [None, "a0", "a0", "a1"], "B": [1, 2, 2, 3]})
        grouped_dataframe = GroupedDataFrame(
            dataframe=self.spark.createDataFrame(data),
            group_keys=self.spark.createDataFrame(
                pd.DataFrame({"A": [None, "a0", "a999"]})
            ),
        )
        expected = pd.DataFrame({"A": [None, "a0", "a999"], "sum(B)": [1, 4, 0]})
        actual = grouped_dataframe.agg(sf.sum("B"), fill_value=0).toPandas()
        self.assert_frame_equal_with_sort(actual, expected)

    def test_apply_in_pandas_with_nulls(self) -> None:
        """Test that .apply_in_pandas works correctly with nulls."""
        data = pd.DataFrame({"A": [None, "a0", "a0", "a1"], "B": [1, 2, 2, 3]})
        grouped_dataframe = GroupedDataFrame(
            dataframe=self.spark.createDataFrame(data),
            group_keys=self.spark.createDataFrame(
                pd.DataFrame({"A": [None, "a0", "a999"]})
            ),
        )
        expected = pd.DataFrame({"A": [None, "a0", "a999"], "sum(B)": [1, 4, 0]})
        actual = grouped_dataframe.apply_in_pandas(
            lambda df: pd.DataFrame({"sum(B)": [df["B"].sum()]}),
            StructType([StructField("sum(B)", IntegerType())]),
        ).toPandas()
        self.assert_frame_equal_with_sort(actual, expected)

    def test_agg_from_same_source(self) -> None:
        """Previous implementations would fail when using the same source twice."""
        data = self.spark.createDataFrame(
            [("0", 1), ("1", 0), ("1", 2)], schema=["A", "B"]
        )
        group_keys = data.filter("B = 2")
        grouped_dataframe = GroupedDataFrame(dataframe=data, group_keys=group_keys)
        expected = pd.DataFrame({"A": ["1"], "B": [2], "count(1)": [1]})
        actual = grouped_dataframe.agg(sf.count("*"), fill_value=0).toPandas()
        self.assert_frame_equal_with_sort(actual, expected)

    def test_apply_in_pandas_from_same_source(self) -> None:
        """Previous implementations would fail when using the same source twice."""
        data = self.spark.createDataFrame(
            [("0", 1), ("1", 0), ("1", 2)], schema=["A", "B"]
        )
        group_keys = data.filter("B = 2")
        grouped_dataframe = GroupedDataFrame(dataframe=data, group_keys=group_keys)
        expected = pd.DataFrame({"A": ["1"], "B": [2], "count": [1]})
        actual = grouped_dataframe.apply_in_pandas(
            lambda df: pd.DataFrame({"count": [len(df)]}),
            StructType([StructField("count", IntegerType())]),
        ).toPandas()
        self.assert_frame_equal_with_sort(actual, expected)

    def test_group_keys_no_rows_one_column(self):
        """Tests that group keys must have no columns it is empty."""

        with self.assertRaisesRegex(
            ValueError, "Group keys cannot have no rows, unless it also has no columns"
        ):
            GroupedDataFrame(
                dataframe=self.spark.createDataFrame(
                    pd.DataFrame([(1, 2)], columns=["A", "Z"])
                ),
                group_keys=self.spark.createDataFrame(
                    [], StructType([StructField("A", IntegerType())])
                ),
            )

    @parameterized.expand(
        [
            (
                pd.DataFrame([("A", 1), ("B", 1), ("B", 2)], columns=["X", "Y"]),
                pd.DataFrame([("A",), ("B",)], columns=["X"]),
                pd.DataFrame([("A", 1), ("B", 2)], columns=["X", "count"]),
            ),
            (
                pd.DataFrame([("A", 1), ("B", 1), ("B", 2)], columns=["X", "Y"]),
                pd.DataFrame([("A", 1), ("B", 1)], columns=["X", "Y"]),
                pd.DataFrame([("A", 1, 1), ("B", 1, 1)], columns=["X", "Y", "count"]),
            ),
        ]
    )
    def test_count_agg(
        self, df: pd.DataFrame, group_keys: pd.DataFrame, expected: pd.DataFrame
    ):
        """Tests that count aggregation works on GroupedDataFrames."""
        actual = (
            GroupedDataFrame(
                dataframe=self.spark.createDataFrame(df),
                group_keys=self.spark.createDataFrame(group_keys),
            )
            .agg(sf.count("*").alias("count"), fill_value=0)
            .toPandas()
        )
        self.assert_frame_equal_with_sort(actual, expected)

    @parameterized.expand(
        [
            (
                pd.DataFrame([("A", 1), ("B", 4), ("B", 5)], columns=["X", "Y"]),
                pd.DataFrame([("A",), ("B",)], columns=["X"]),
                pd.DataFrame([("A", 1), ("B", 9)], columns=["X", "sum(Y)"]),
            ),
            (
                pd.DataFrame([("A", 1), ("B", 4), ("C", 5)], columns=["X", "Y"]),
                pd.DataFrame([("A",), ("B",)], columns=["X"]),
                pd.DataFrame([("A", 1), ("B", 4)], columns=["X", "sum(Y)"]),
            ),
            (
                pd.DataFrame([("A", 1), ("B", 4)], columns=["X", "Y"]),
                pd.DataFrame([("A",), ("B",), ("C",)], columns=["X"]),
                pd.DataFrame([("A", 1), ("B", 4), ("C", 0)], columns=["X", "sum(Y)"]),
            ),
        ]
    )
    def test_sum_agg(
        self, df: pd.DataFrame, group_keys: pd.DataFrame, expected: pd.DataFrame
    ):
        """Tests that agg works as expected."""
        sum_func = sf.sum(sf.col("Y")).alias("sum(Y)")
        self.assert_frame_equal_with_sort(
            GroupedDataFrame(
                dataframe=self.spark.createDataFrame(df),
                group_keys=self.spark.createDataFrame(group_keys),
            )
            .agg(sum_func, fill_value=0)
            .toPandas(),
            expected,
        )

    @parameterized.expand(
        [
            (
                pd.DataFrame([("A", 1), ("B", 4), ("B", 5)], columns=["X", "Y"]),
                pd.DataFrame([("A",), ("B",)], columns=["X"]),
                pd.DataFrame([("A", 1), ("B", 9)], columns=["X", "sum(Y)"]),
            ),
            (
                pd.DataFrame([("A", 1), ("B", 4), ("C", 5)], columns=["X", "Y"]),
                pd.DataFrame([("A",), ("B",)], columns=["X"]),
                pd.DataFrame([("A", 1), ("B", 4)], columns=["X", "sum(Y)"]),
            ),
            (
                pd.DataFrame([("A", 1), ("B", 4)], columns=["X", "Y"]),
                pd.DataFrame([("A",), ("B",), ("C",)], columns=["X"]),
                pd.DataFrame([("A", 1), ("B", 4), ("C", 0)], columns=["X", "sum(Y)"]),
            ),
        ]
    )
    def test_sum_apply_in_pandas(
        self, df: pd.DataFrame, group_keys: pd.DataFrame, expected: pd.DataFrame
    ):
        """Tests that apply_in_pandas works as expected."""
        self.assert_frame_equal_with_sort(
            GroupedDataFrame(
                dataframe=self.spark.createDataFrame(df),
                group_keys=self.spark.createDataFrame(group_keys),
            )
            .apply_in_pandas(
                lambda df: pd.DataFrame({"sum(Y)": [df["Y"].sum()]}),
                StructType([StructField("sum(Y)", IntegerType())]),
            )
            .toPandas(),
            expected,
        )

    def test_empty_agg(self):
        """Tests that agg works for empty group keys."""
        sum_func = sf.sum(sf.col("Y")).alias("sum(Y)")
        self.assert_frame_equal_with_sort(
            GroupedDataFrame(
                dataframe=self.spark.createDataFrame(
                    pd.DataFrame([("A", 1), ("B", 4)], columns=["X", "Y"])
                ),
                group_keys=self.spark.createDataFrame([], schema=StructType()),
            )
            .agg(sum_func, fill_value=0)
            .toPandas(),
            pd.DataFrame({"sum(Y)": [5]}),
        )

    def test_empty_apply_in_pandas(self):
        """Tests that apply_in_pandas works for empty group keys."""
        grouped_dataframe = GroupedDataFrame(
            dataframe=self.spark.createDataFrame(
                pd.DataFrame([("A", 1), ("B", 4)], columns=["X", "Y"])
            ),
            group_keys=self.spark.createDataFrame([], schema=StructType()),
        )
        actual = grouped_dataframe.apply_in_pandas(
            lambda df: pd.DataFrame({"sum(Y)": [df["Y"].sum()]}),
            StructType([StructField("sum(Y)", IntegerType())]),
        ).toPandas()
        expected = pd.DataFrame({"sum(Y)": [5]})
        self.assert_frame_equal_with_sort(actual, expected)

    def test_agg_fill_value(self):
        """Tests that agg fills correct value for missing keys."""
        expected = pd.DataFrame({"X": ["A", "B"], "sum(Y)": [1, 10]})
        actual = (
            GroupedDataFrame(
                dataframe=self.spark.createDataFrame([("A", 1)], schema=["X", "Y"]),
                group_keys=self.spark.createDataFrame([("A",), ("B",)], schema=["X"]),
            )
            .agg(func=sf.sum(sf.col("Y")).alias("sum(Y)"), fill_value=10)
            .toPandas()
        )
        self.assert_frame_equal_with_sort(expected, actual)

    def test_agg_does_not_override_valid_nulls(self):
        """Tests that agg does not replace nulls associated with existing keys."""
        expected = pd.DataFrame({"X": ["A", "B", "C"], "sum(Y)": [1, None, 0]})
        actual = (
            GroupedDataFrame(
                dataframe=self.spark.createDataFrame(
                    [("A", 1), ("B", None)], schema=["X", "Y"]
                ),
                group_keys=self.spark.createDataFrame(
                    [("A",), ("B",), ("C",)], schema=["X"]
                ),
            )
            .agg(func=sf.sum(sf.col("Y")).alias("sum(Y)"), fill_value=0)
            .toPandas()
        )
        self.assert_frame_equal_with_sort(expected, actual)

    def test_get_groups(self):
        """Tests `get_groups` returns correct groups."""
        actual_groups = GroupedDataFrame(
            dataframe=self.spark.createDataFrame(
                [("A", 1), ("B", 2), ("B", 4), (None, 2), (None, 3)], schema=["X", "Y"]
            ),
            group_keys=self.spark.createDataFrame(
                [("A",), ("B",), ("C",), (None,)], schema=["X"]
            ),
        ).get_groups()
        expected_groups = {
            "A": pd.DataFrame({"Y": [1]}),
            "B": pd.DataFrame({"Y": [2, 4]}),
            "C": pd.DataFrame({"Y": []}),
            None: pd.DataFrame({"Y": [2, 3]}),
        }
        for key, expected_group in expected_groups.items():
            self.assert_frame_equal_with_sort(
                expected_group, actual_groups[Row(A=key)].toPandas()
            )



"""Test for :mod:`tmlt.core.utils.arb`"""


# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

import math
from unittest import TestCase

from parameterized import parameterized
from scipy.special import erf, erfc  # pylint: disable=no-name-in-module

from tmlt.core.utils.arb import (
    Arb,
    arb_abs,
    arb_add,
    arb_const_pi,
    arb_div,
    arb_erf,
    arb_erfc,
    arb_erfinv,
    arb_exp,
    arb_log,
    arb_max,
    arb_min,
    arb_mul,
    arb_neg,
    arb_product,
    arb_sgn,
    arb_sub,
    arb_sum,
    arb_union,
)


class TestArb(TestCase):
    """Tests for :class:`tmlt.core.utils.arb.Arb`."""

    @parameterized.expand(
        [
            (-42 * 10**9,),
            (-42 * 10**7,),
            (-42,),
            (0,),
            (42,),
            (42 * 10**7,),
            (42 * 10**9,),
            (42 * 10**11,),
        ]
    )
    def test_from_int(self, val: int):
        """Tests `Arb.from_int`."""
        x = Arb.from_int(val)
        man, exp = x.man_exp()
        self.assertEqual(man * 2**exp, val)

    @parameterized.expand([(0.0,), (1.1,), (-1.1,), (9.9 * 0.123,), (99.12,)])
    def test_from_float(self, val: float):
        """Tests `Arb.from_float`."""
        x = Arb.from_float(val)
        man, exp = x.man_exp()
        self.assertEqual(man * 2**exp, val)

    def test_from_float_inf(self):
        """Tests `Arb.from_float` works with +/- inf."""
        posinf = Arb.from_float(float("inf"))
        neginf = Arb.from_float(float("-inf"))

        self.assertFalse(posinf.is_finite())
        self.assertFalse(neginf.is_finite())
        self.assertEqual(float(posinf), float("inf"))
        self.assertEqual(float(neginf), float("-inf"))

    @parameterized.expand([(2, 30), (4, 300), (5 * 10**2, 7**8)])
    def test_from_man_exp(self, man: int, exp: int):
        """Tests `Arb.from_man_exp`."""
        x = Arb.from_man_exp(man, exp)
        m, e = x.man_exp()
        self.assertEqual(m * 2**e, man * 2**exp)

    @parameterized.expand([(10, 1), (10000, 5), (10, 1), (10, 1)])
    def test_from_midpoint_radius(self, mid: int, rad: int):
        """Tests `Arb.from_midpoint_radius`."""
        mid_arb = Arb.from_int(mid)
        rad_arb = Arb.from_int(rad)
        x = Arb.from_midpoint_radius(mid_arb, rad_arb)
        assert x.midpoint() == mid_arb
        actual_radius = float(x.radius())
        self.assertAlmostEqual(actual_radius, rad)

    @parameterized.expand(
        [
            (Arb.from_int(10), True),
            (Arb.from_float(0.01), True),
            (Arb.from_float(-float("inf")), True),
            (Arb.from_midpoint_radius(1, 0), True),
            (Arb.from_midpoint_radius(1, 1), False),
        ]
    )
    def test_is_exact(self, arb: Arb, exact: bool):
        """Tests `Arb.is_exact`."""
        self.assertEqual(arb.is_exact(), exact)

    def test_is_finite(self):
        """Tests `Arb.is_finite`."""
        self.assertFalse(Arb.from_float(-float("inf")).is_finite())
        self.assertFalse(Arb.from_float(float("inf")).is_finite())
        self.assertTrue(Arb.from_int(10).is_finite())

    def test_is_nan(self):
        """Tests `Arb.is_nan`."""
        self.assertTrue(Arb.from_float(float("nan")).is_nan())
        self.assertFalse(Arb.from_float(0.0).is_nan())

    def test_lower(self):
        """Tests `Arb.lower`."""
        arb = Arb.from_midpoint_radius(1, 0.5)
        self.assertAlmostEqual(float(arb.lower(prec=100)), 0.5)

    def test_upper(self):
        """Tests `Arb.upper`."""
        arb = Arb.from_midpoint_radius(1, 0.5)
        self.assertAlmostEqual(float(arb.upper(prec=100)), 1.5)

    @parameterized.expand(
        [
            (
                Arb.from_midpoint_radius(mid=9, rad=1),
                Arb.from_midpoint_radius(mid=10, rad=2),
                True,
            ),
            (
                Arb.from_midpoint_radius(mid=10, rad=2),
                Arb.from_midpoint_radius(mid=9, rad=1),
                False,
            ),
            (Arb.from_int(10), Arb.from_midpoint_radius(mid=9, rad=1), True),
            (Arb.from_float(10.1), Arb.from_midpoint_radius(mid=9, rad=1), False),
        ]
    )
    def test_contains(self, x: Arb, y: Arb, expected: bool):
        """`y.__contains__(x)` returns True iff every number in `x` is also in `y`."""
        self.assertEqual(x in y, expected)

    @parameterized.expand(
        [
            (Arb.from_int(10), Arb.from_int(10), True),
            (Arb.from_int(10), Arb.from_int(11), False),
            (Arb.from_float(10.0), Arb.from_int(10), True),
            (
                Arb.from_midpoint_radius(mid=10, rad=2),
                Arb.from_midpoint_radius(mid=10, rad=2),
                True,
            ),
            (
                Arb.from_midpoint_radius(mid=10, rad=2),
                Arb.from_midpoint_radius(mid=10, rad=3),
                False,
            ),
            (arb_const_pi(100), arb_const_pi(100), True),
            (arb_const_pi(100), arb_const_pi(1000), False),
        ]
    )
    def test_hash(self, x: Arb, y: Arb, expected: bool):
        """`x` and `y` hash to the same value if they have the same midpoint and radius.

        Args:
            x: An Arb.
            y: An Arb.
            expected: Whether `x` and `y` should hash to the same value.
        """
        self.assertEqual(hash(x) == hash(y), expected)


class TestArbFunctions(TestCase):
    """Tests for arithmetic functions in :mod:`tmlt.core.utils.arb`.

    NOTE: Correctness of an arb function `F` is specified as follows:

        If `f` is the corresponding real-valued arithmetic function, `F` is correct
        only if, for any Arb X and any real number x in the interval X,
        `f(x)` is in `F(X)`.

    These tests assume Arb.__contains__ is correct.
    """

    def test_arb_sub(self):
        """`arb_sub` works as expected."""
        arb1 = Arb.from_midpoint_radius(2, 0.5)
        arb2 = Arb.from_midpoint_radius(1, 1)
        actual = arb_sub(arb1, arb2, prec=100)
        # Smallest value in diff => 1.5 - 2 = -0.5
        # Largest value in diff => 2.5 - 0 = 2.5
        true_interval = Arb.from_midpoint_radius(1, 1.5)  # [-0.5, 2.5]
        self.assertTrue(true_interval in actual)

    def test_arb_add(self):
        """`arb_add` works as expected."""
        arb1 = Arb.from_midpoint_radius(2, 1)
        arb2 = Arb.from_midpoint_radius(1, 1)
        actual = arb_add(arb1, arb2, prec=100)
        true_interval = Arb.from_midpoint_radius(3, 2)  # [1, 5]
        self.assertTrue(true_interval in actual)

    def test_arb_mul(self):
        """`arb_mul` works as expected."""
        arb1 = Arb.from_midpoint_radius(2, 1)
        arb2 = Arb.from_midpoint_radius(1, 1)
        actual = arb_mul(arb1, arb2, prec=100)
        true_interval = Arb.from_midpoint_radius(3, 3)  # [0, 6]
        self.assertTrue(true_interval in actual)

    def test_arb_div(self):
        """`arb_div` works as expected."""
        arb1 = Arb.from_midpoint_radius(4, 1)
        arb2 = Arb.from_midpoint_radius(2, 1)
        actual = arb_div(arb1, arb2, prec=100)
        true_interval = Arb.from_midpoint_radius(4, 1)  # [3, 5]
        self.assertTrue(true_interval in actual)

    def test_arb_log(self):
        """`arb_log` works as expected."""
        midpoint = (1 + math.exp(10)) / 2
        arb = Arb.from_midpoint_radius(midpoint, midpoint - 1)  # [1, exp(10)]
        actual = arb_log(arb, prec=100)
        true_interval = Arb.from_midpoint_radius(5, 5)  # [0,10]
        self.assertTrue(true_interval in actual)

    def test_arb_exp(self):
        """`arb_exp` works as expected."""
        midpoint = math.log(9) / 2
        arb = Arb.from_midpoint_radius(midpoint, midpoint)  # [0, log(9)]
        actual = arb_exp(arb, prec=100)
        true_interval = Arb.from_midpoint_radius(5, 4)  # [1,9]
        self.assertTrue(true_interval in actual)

    def test_arb_max(self):
        """`arb_max` works as expected."""
        arb1 = Arb.from_midpoint_radius(1.5, 0.5)  # [1, 2]
        arb2 = Arb.from_midpoint_radius(1, 2)  # [-1, 3]
        actual = arb_max(arb1, arb2, prec=100)
        true_interval = Arb.from_midpoint_radius(2, 1)  # [1, 3]
        self.assertTrue(true_interval in actual)

    def test_arb_min(self):
        """`arb_min` works as expected."""
        arb1 = Arb.from_midpoint_radius(1.5, 0.5)  # [1, 2]
        arb2 = Arb.from_midpoint_radius(1, 2)  # [-1, 3]
        actual = arb_min(arb1, arb2, prec=100)
        true_interval = Arb.from_midpoint_radius(0.5, 1.5)  # [-1, 2]
        self.assertTrue(true_interval in actual)

    def test_arb_abs(self):
        """`arb_abs` works as expected."""
        arb = Arb.from_midpoint_radius(1, 2)  # [-1,3]
        actual = arb_abs(arb)
        true_interval = Arb.from_midpoint_radius(1.5, 1.5)
        self.assertTrue(true_interval in actual)

    def test_arb_neg(self):
        """`arb_neg` works as expected."""
        arb = Arb.from_midpoint_radius(1, 2)  # [-1,3]
        actual = arb_neg(arb)
        true_interval = Arb.from_midpoint_radius(-2, 1)  # [-3,1]
        self.assertTrue(true_interval in actual)

    def test_arb_sgn(self):
        """`arb_sgn` works as expected."""
        arb1 = Arb.from_midpoint_radius(1, 0.5)  # [0.5,1.5]
        arb2 = Arb.from_midpoint_radius(-1, 0.5)  # [-1.5,-0.5]
        arb3 = Arb.from_midpoint_radius(1, 2)  # [-1,3]
        self.assertAlmostEqual(arb_sgn(arb1).to_float(), 1)
        self.assertAlmostEqual(arb_sgn(arb2).to_float(), -1)
        # arb3 contains both positive and negative numbers
        # So, arb_sgn returns [0, 1]
        self.assertAlmostEqual(float(arb_sgn(arb3).midpoint()), 0)
        self.assertAlmostEqual(float(arb_sgn(arb3).radius()), 1)

    def test_arb_erfinv(self):
        """`arb_erfinv` works as expected."""
        midpoint = (erf(1 / 8) + erf(1 / 16)) / 2
        radius = midpoint - erf(1 / 16)
        arb = Arb.from_midpoint_radius(midpoint, radius)
        actual = arb_erfinv(arb, prec=100)
        true_interval = Arb.from_midpoint_radius(3 / 32, 1 / 32)  # [1/16, 1/8]
        self.assertTrue(true_interval in actual)

    def test_arb_erf(self):
        """`arb_erf` works as expected."""
        arb = Arb.from_midpoint_radius(2, 1)
        actual = arb_erf(arb, prec=100)
        true_interval = Arb.from_midpoint_radius(
            (erf(1) + erf(3)) / 2, (erf(1) + erf(3)) / 2 - erf(1)
        )
        self.assertTrue(true_interval in actual)

    def test_arb_erfc(self):
        """`arb_erfc` works as expected."""
        arb = Arb.from_midpoint_radius(2, 1)
        actual = arb_erfc(arb, prec=100)
        true_interval = Arb.from_midpoint_radius(
            (erfc(1) + erfc(3)) / 2, (erfc(1) + erfc(3)) / 2 - erfc(3)
        )
        self.assertTrue(true_interval in actual)

    def test_arb_const_pi(self):
        """`arb_const_pi` works as expected."""
        actual = arb_const_pi(100)
        interval_around_pi = Arb.from_midpoint_radius(math.pi, 1e-10)
        self.assertTrue(actual in interval_around_pi)

    def test_arb_union(self):
        """`arb_union` works as expected."""
        arb1 = Arb.from_midpoint_radius(1, 0.5)  # [0.5,1.5]
        arb2 = Arb.from_midpoint_radius(3, 0.5)  # [2.5,3.5]
        actual = arb_union(arb1, arb2)
        true_interval = Arb.from_midpoint_radius(2, 1.5)  # [0.5, 3.5]
        self.assertTrue(true_interval in actual)

    def test_arb_sum(self):
        """`arb_sum` works as expected."""
        arb1 = Arb.from_midpoint_radius(1, 0.5)  # [0.5,1.5]
        arb2 = Arb.from_midpoint_radius(2, 0.5)  # [1.5,2.5]
        arb3 = Arb.from_midpoint_radius(3, 0.5)  # [2.5,3.5]
        actual = arb_sum([arb1, arb2, arb3], prec=100)
        true_interval = Arb.from_midpoint_radius(6, 1.5)  # [4.5, 7.5]
        self.assertTrue(true_interval in actual)

    def test_arb_product(self):
        """`arb_product` works as expected."""
        arb1 = Arb.from_midpoint_radius(1, 0.5)  # [0.5,1.5]
        arb2 = Arb.from_midpoint_radius(2, 1)  # [1,3]
        arb3 = Arb.from_midpoint_radius(0, 1)  # [-1,1]
        actual = arb_product([arb1, arb2, arb3], prec=100)
        true_interval = Arb.from_midpoint_radius(0, 4.5)  # [-4.5, 4.5]
        self.assertTrue(true_interval in actual)



"""Unit tests for :mod:`~tmlt.core.domains.spark_domains`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

import copy
import datetime
from collections.abc import Mapping
from contextlib import nullcontext as does_not_raise
from itertools import combinations_with_replacement, product
from test.conftest import assert_frame_equal_with_sort
from test.unit.domains.abstract import DomainTests
from typing import Any, Callable, ContextManager, Dict, List, Optional, Type

import pandas as pd
import pytest
from pyspark import Row
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import (
    DataType,
    DateType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

from tmlt.core.domains.base import Domain, OutOfDomainError
from tmlt.core.domains.collections import DictDomain, ListDomain
from tmlt.core.domains.numpy_domains import (
    NumpyDomain,
    NumpyFloatDomain,
    NumpyIntegerDomain,
    NumpyStringDomain,
)
from tmlt.core.domains.pandas_domains import PandasDataFrameDomain, PandasSeriesDomain
from tmlt.core.domains.spark_domains import (
    SparkColumnDescriptor,
    SparkColumnsDescriptor,
    SparkDataFrameDomain,
    SparkDateColumnDescriptor,
    SparkFloatColumnDescriptor,
    SparkGroupedDataFrameDomain,
    SparkIntegerColumnDescriptor,
    SparkRowDomain,
    SparkStringColumnDescriptor,
    SparkTimestampColumnDescriptor,
    convert_numpy_domain,
    convert_pandas_domain,
    convert_spark_schema,
)
from tmlt.core.utils.grouped_dataframe import GroupedDataFrame
from tmlt.core.utils.misc import get_fullname
from tmlt.core.utils.testing import get_all_props


@pytest.mark.usefixtures("class_spark")
class TestSparkDataFrameDomain(DomainTests):
    """Tests for :class:`~tmlt.core.domains.spark_domains.SparkDataFrameDomain`."""

    spark: SparkSession

    @pytest.fixture
    def domain_type(self) -> Type[Domain]:  # pylint: disable=no-self-use
        """Returns the type of the domain to be tested."""
        return SparkDataFrameDomain

    @pytest.fixture(scope="class")
    def domain(self) -> SparkDataFrameDomain:  # pylint: disable=no-self-use
        """Get a base SparkDataFrameDomain."""
        return SparkDataFrameDomain(
            schema={
                "A": SparkStringColumnDescriptor(),
                "B": SparkStringColumnDescriptor(),
                "C": SparkFloatColumnDescriptor(),
            }
        )

    @pytest.mark.parametrize(
        "domain_args, expectation, exception_properties",
        [
            (
                {"schema": invalid_schema},
                pytest.raises(
                    TypeError,
                    match=f'type of argument "schema" must be {get_fullname(Mapping)}; '
                    f"got {get_fullname(invalid_schema)} instead",
                ),
                None,
            )
            for invalid_schema in [StringType, ListDomain(NumpyIntegerDomain())]
        ]
        + [
            (
                {"schema": {"A": SparkStringColumnDescriptor(), "B": DictDomain({})}},
                pytest.raises(
                    TypeError,
                    match=f"Expected domain for key 'B' to be a "
                    f"{get_fullname(SparkColumnDescriptor)}; got "
                    f"{get_fullname(DictDomain)} instead",
                ),
                None,
            ),
            (
                {"schema": {"A": "B"}},
                pytest.raises(
                    TypeError,
                    match=f"Expected domain for key 'A' to be a "
                    f"{get_fullname(SparkColumnDescriptor)}; got "
                    f"{get_fullname(str)} instead",
                ),
                None,
            ),
        ]
        + [
            ({"schema": valid_schema}, does_not_raise(), None)
            for valid_schema in [
                {
                    "A": SparkStringColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                },
                {},
                {"A": SparkStringColumnDescriptor()},
            ]
        ],
    )
    def test_construct_component(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Initialization behaves correctly.

        The domain is constructed correctly and raises exceptions when initialized with
        invalid inputs.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        super().test_construct_component(
            domain_type, domain_args, expectation, exception_properties
        )

    @pytest.mark.parametrize(
        "other_domain, expected",
        [
            (  # matching
                SparkDataFrameDomain(
                    {
                        "A": SparkStringColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(),
                    }
                ),
                True,
            ),
            (  # shuffled
                SparkDataFrameDomain(
                    {
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(),
                        "A": SparkStringColumnDescriptor(),
                    }
                ),
                False,
            ),
            (  # Mismatching Types
                SparkDataFrameDomain(
                    {
                        "A": SparkStringColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(size=32),
                    }
                ),
                False,
            ),
            (  # Extra attribute
                SparkDataFrameDomain(
                    {
                        "A": SparkStringColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(),
                        "D": SparkFloatColumnDescriptor(),
                    }
                ),
                False,
            ),
            (  # Missing attribute
                SparkDataFrameDomain(
                    {
                        "A": SparkStringColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                ),
                False,
            ),
        ],
    )
    def test_eq(self, domain: Domain, other_domain: Domain, expected: bool):
        """__eq__ works correctly.

        Args:
            domain: The domain to test.
            other_domain: The domain to compare to.
            expected: The expected result of the comparison.
        """
        super().test_eq(domain, other_domain, expected)

    @pytest.mark.parametrize(
        "domain_args, key, mutator",
        [
            (
                {
                    "schema": {
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(),
                    }
                },
                "schema",
                mutator,
            )
            for mutator in [
                lambda x: x.update({"A": SparkFloatColumnDescriptor()}),
                lambda x: x.pop("A"),
                lambda x: x.clear(),
            ]
        ],
    )
    def test_mutable_inputs(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        key: str,
        mutator: Callable[[Any], Any],
    ):
        """The mutable inputs to the domain are copied.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(domain_type, domain_args, key, mutator)

    @pytest.mark.parametrize(
        "domain, expected_properties",
        [
            (
                SparkDataFrameDomain(
                    schema={
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(),
                    }
                ),
                {
                    "schema": {
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(),
                    },
                    "carrier_type": DataFrame,
                    "spark_schema": StructType(
                        [
                            StructField("A", LongType(), False),
                            StructField("B", StringType(), False),
                            StructField("C", DoubleType(), False),
                        ]
                    ),
                },
            ),
            (
                SparkDataFrameDomain(
                    schema={
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(
                            allow_inf=True, allow_nan=True, allow_null=True
                        ),
                    }
                ),
                {
                    "schema": {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(
                            allow_inf=True, allow_nan=True, allow_null=True
                        ),
                    },
                    "carrier_type": DataFrame,
                    "spark_schema": StructType(
                        [
                            StructField("A", LongType(), True),
                            StructField("B", StringType(), True),
                            StructField("C", DoubleType(), True),
                        ]
                    ),
                },
            ),
        ],
    )
    def test_properties(self, domain: Domain, expected_properties: Dict[str, Any]):
        """All properties have the expected values.

        Args:
            domain: The constructed domain to be tested.
            expected_properties: A dictionary containing all the property:value pairs
                domain is expected to have.
        """
        super().test_properties(domain, expected_properties)

    @pytest.mark.parametrize(
        "domain",
        [
            SparkDataFrameDomain(
                schema={
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                    "C": SparkFloatColumnDescriptor(),
                }
            )
        ],
    )
    def test_property_immutability(self, domain: Domain):
        """The properties return copies for mutable values.

        Args:
            domain: The domain to be tested.
        """
        super().test_property_immutability(domain)

    @pytest.mark.parametrize(
        "candidate, expectation, exception_properties",
        [
            (  # LongType() instead of DoubleType()
                pd.DataFrame(
                    [["A", "B", 10], ["V", "E", 12], ["A", "V", 13]],
                    columns=["A", "B", "C"],
                ),
                pytest.raises(
                    OutOfDomainError,
                    match="Found invalid value in column 'C': Column must be "
                    f"{get_fullname(DoubleType)}; got {get_fullname(LongType)} "
                    "instead",
                ),
                {
                    "domain": SparkDataFrameDomain(
                        schema={
                            "A": SparkStringColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                            "C": SparkFloatColumnDescriptor(),
                        }
                    ),
                    "value": pd.DataFrame(
                        [["A", "B", 10], ["V", "E", 12], ["A", "V", 13]],
                        columns=["A", "B", "C"],
                    ),
                },
            ),
            (  # Missing Columns
                pd.DataFrame([["A", "B"], ["V", "E"], ["A", "V"]], columns=["A", "B"]),
                pytest.raises(
                    OutOfDomainError,
                    match="Columns are not as expected. DataFrame and Domain "
                    "must contain the same columns in the same order.\n"
                    r"DataFrame columns: \['A', 'B'\]"
                    "\n"
                    r"Domain columns: \['A', 'B', 'C'\]",
                ),
                {
                    "domain": SparkDataFrameDomain(
                        schema={
                            "A": SparkStringColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                            "C": SparkFloatColumnDescriptor(),
                        }
                    ),
                    "value": pd.DataFrame(
                        [["A", "B"], ["V", "E"], ["A", "V"]], columns=["A", "B"]
                    ),
                },
            ),
            (
                pd.DataFrame(
                    [["A", "B", 1.1], ["V", "E", 1.2], ["A", "V", None]],
                    columns=["A", "B", "C"],
                ),
                pytest.raises(
                    OutOfDomainError,
                    match="Found invalid value in column 'C': Column contains null "
                    "values.",
                ),
                {
                    "domain": SparkDataFrameDomain(
                        schema={
                            "A": SparkStringColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                            "C": SparkFloatColumnDescriptor(),
                        }
                    ),
                    "value": pd.DataFrame(
                        [["A", "B", 1.1], ["V", "E", 1.2], ["A", "V", None]],
                        columns=["A", "B", "C"],
                    ),
                },
            ),
            (
                pd.DataFrame(
                    [["A", "B", 1.1], ["V", "E", 1.2], ["A", "V", 1.3]],
                    columns=["A", "B", "C"],
                ),
                does_not_raise(),
                None,
            ),
        ],
    )
    def test_validate(
        self,
        domain: Domain,
        candidate: Any,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Validate works correctly.

        Args:
            domain: The domain to test.
            candidate: The value to validate using domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        if isinstance(candidate, pd.DataFrame):
            candidate = self.spark.createDataFrame(candidate)
        with expectation as exception:
            domain.validate(candidate)
        if exception_properties is None or len(exception_properties) == 0:
            return
        # Help out mypy
        assert isinstance(exception, pytest.ExceptionInfo)
        for prop, expected_value in exception_properties.items():
            assert hasattr(exception.value, prop), f"Expected prop was missing: {prop}"
            actual_value = getattr(exception.value, prop)
            if isinstance(actual_value, DataFrame):
                assert_frame_equal_with_sort(actual_value, expected_value)
                continue
            assert (
                actual_value == expected_value
            ), f"Expected {prop} to be {expected_value}, got {actual_value}"

    @pytest.mark.parametrize(
        "spark_schema, expected, expectation",
        [
            (StructType([]), SparkDataFrameDomain(schema={}), does_not_raise()),
            (
                StructType(
                    [
                        StructField("A", LongType(), False),
                        StructField("B", StringType(), False),
                        StructField("C", DoubleType(), False),
                    ]
                ),
                SparkDataFrameDomain(
                    schema={
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(allow_inf=True, allow_nan=True),
                    }
                ),
                does_not_raise(),
            ),
        ],
    )
    def test_from_spark_schema(  # pylint: disable=no-self-use
        self,
        spark_schema: StructType,
        expected: SparkDataFrameDomain,
        expectation: ContextManager[None],
    ):
        """from_spark_schema constructs the correct domain.

        Args:
            spark_schema: The spark schema to construct the domain from.
            expected: The expected domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        with expectation:
            assert SparkDataFrameDomain.from_spark_schema(spark_schema) == expected


_base_schema: Dict[str, SparkColumnDescriptor] = {
    "A": SparkIntegerColumnDescriptor(allow_null=True),
    "B": SparkStringColumnDescriptor(allow_null=True),
    "C": SparkIntegerColumnDescriptor(allow_null=True),
}
_schema_without_nulls: Dict[str, SparkColumnDescriptor] = {
    "A": SparkIntegerColumnDescriptor(allow_null=False),
    "B": SparkStringColumnDescriptor(allow_null=False),
    "C": SparkIntegerColumnDescriptor(allow_null=False),
}

_base_groupby_columns: List[str] = ["A", "B"]
_base_group_key_args: Dict[str, Any] = {
    "data": [(1, "W"), (2, "X"), (3, "Y")],
    "schema": ["A", "B"],
}
_group_key_with_nulls_args: Dict[str, Any] = {
    "data": [(1, "W"), (2, "X"), (3, None)],
    "schema": ["A", "B"],
}
_empty_group_key_args: Dict[str, Any] = {"data": [], "schema": StructType([])}


# Helper functions to help mypy out by widening the type of a context manager.
def _WidenContextManager(cm: Any) -> ContextManager:
    """Widen the type of a context manager."""
    return cm


@pytest.mark.usefixtures("class_spark")
class TestSparkGroupedDataFrameDomain(DomainTests):
    """Testing :class:`~tmlt.core.domains.spark_domains.SparkGroupedDataFrameDomain`."""

    spark: SparkSession

    @pytest.fixture
    def domain_type(self) -> Type[Domain]:  # pylint: disable=no-self-use
        """Returns the type of the domain to be tested."""
        return SparkGroupedDataFrameDomain

    @pytest.mark.parametrize(
        "domain_args, expectation, exception_properties",
        [
            (
                {"schema": _base_schema, "groupby_columns": _base_groupby_columns},
                does_not_raise(),
                None,
            ),
            # _base_schema does not have column "D"
            (
                {"schema": _base_schema, "groupby_columns": ["D"]},
                pytest.raises(ValueError, match="Invalid groupby columns: {'D'}"),
                None,
            ),
            # Invalid schema
            (
                {"schema": "not a schema", "groupby_columns": ["C"]},
                pytest.raises(
                    TypeError,
                    match=f'type of argument "schema" must be {get_fullname(Mapping)}; '
                    f"got {get_fullname(str)} instead",
                ),
                None,
            ),
            # Invalid Column "C"
            (
                {
                    "schema": {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": ListDomain(NumpyIntegerDomain()),
                    },
                    "groupby_columns": ["A"],
                },
                pytest.raises(
                    TypeError,
                    match=f"Expected domain for key 'C' to be a "
                    f"{get_fullname(SparkColumnDescriptor)}; got "
                    f"{get_fullname(ListDomain)} instead",
                ),
                None,
            ),
        ],
    )
    def test_construct_component(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Initialization behaves correctly.

        The domain is constructed correctly and raises exceptions when initialized with
        invalid inputs.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        super().test_construct_component(
            domain_type, domain_args, expectation, exception_properties
        )

    @pytest.mark.parametrize(
        "domain, other_domain, expected",
        [
            (
                # eq with same schema and groupby_columns
                SparkGroupedDataFrameDomain(
                    schema=_base_schema, groupby_columns=_base_groupby_columns
                ),
                SparkGroupedDataFrameDomain(
                    schema=_base_schema, groupby_columns=_base_groupby_columns
                ),
                True,
            ),
            (
                # eq with no groupby columns
                SparkGroupedDataFrameDomain(schema=_base_schema, groupby_columns=[]),
                SparkGroupedDataFrameDomain(schema=_base_schema, groupby_columns=[]),
                True,
            ),
            (
                # not eq with different schemas
                SparkGroupedDataFrameDomain(
                    schema=_base_schema, groupby_columns=_base_groupby_columns
                ),
                SparkGroupedDataFrameDomain(
                    schema=_schema_without_nulls, groupby_columns=_base_groupby_columns
                ),
                False,
            ),
        ],
    )
    def test_eq(self, domain: Domain, other_domain: Domain, expected: bool):
        """__eq__ works correctly.

        Args:
            domain: The domain to test.
            other_domain: The domain to compare to.
            expected: The expected result of the comparison.
        """
        super().test_eq(domain, other_domain, expected)

    @pytest.mark.parametrize(
        "domain_args, key, mutator",
        [
            (
                {
                    "schema": copy.deepcopy(_base_schema),
                    "groupby_columns": copy.deepcopy(_base_groupby_columns),
                },
                "schema",
                lambda x: x.update({"A": SparkFloatColumnDescriptor()}),
            )
        ],
    )
    def test_mutable_inputs(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        key: str,
        mutator: Callable[[Any], Any],
    ):
        """The mutable inputs to the domain are copied.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(domain_type, domain_args, key, mutator)

    @pytest.mark.parametrize(
        "domain, expected_properties",
        [
            (
                SparkGroupedDataFrameDomain(_base_schema, _base_groupby_columns),
                {
                    "schema": _base_schema,
                    "carrier_type": GroupedDataFrame,
                    "spark_schema": StructType(
                        [
                            StructField("A", LongType(), True),
                            StructField("B", StringType(), True),
                            StructField("C", LongType(), True),
                        ]
                    ),
                    "groupby_columns": _base_groupby_columns,
                },
            )
        ],
    )
    def test_properties(self, domain: Domain, expected_properties: Dict[str, Any]):
        """All properties have the expected values.

        Args:
            domain: The constructed domain to be tested.
            expected_properties: A dictionary containing all the property:value pairs
                domain is expected to have.
        """
        actual_props = [prop[0] for prop in get_all_props(type(domain))]
        assert set(expected_properties.keys()) == set(actual_props)
        for prop, expected_val in expected_properties.items():
            assert hasattr(domain, prop) and getattr(domain, prop) == expected_val

    @pytest.mark.parametrize(
        "domain", [SparkGroupedDataFrameDomain(_base_schema, _base_groupby_columns)]
    )
    def test_property_immutability(self, domain: Domain):
        """The properties return copies for mutable values.

        Args:
            domain: The domain to be tested.
        """
        super().test_property_immutability(domain)

    @pytest.mark.parametrize(
        "domain, candidate, expectation, exception_properties",
        [
            (
                # Normal
                SparkGroupedDataFrameDomain(_base_schema, _base_groupby_columns),
                {
                    "dataframe": {
                        "data": [(1, "W", 10), (2, "X", 12), (3, "Y", 13)],
                        "schema": ["A", "B", "C"],
                    },
                    "group_keys": _base_group_key_args,
                },
                does_not_raise(),
                None,
            ),
            (
                # Nulls
                SparkGroupedDataFrameDomain(_base_schema, _base_groupby_columns),
                {
                    "dataframe": {
                        "data": [(1, "W", 10), (2, "X", 12), (3, "Y", None)],
                        "schema": ["A", "B", "C"],
                    },
                    "group_keys": _group_key_with_nulls_args,
                },
                does_not_raise(),
                None,
            ),
            (
                # Unexpected nulls in dataframe
                SparkGroupedDataFrameDomain(
                    _schema_without_nulls, _base_groupby_columns
                ),
                {
                    "dataframe": {
                        "data": [(1, "W", 10), (2, "X", 12), (3, "Y", None)],
                        "schema": ["A", "B", "C"],
                    },
                    "group_keys": _base_group_key_args,
                },
                pytest.raises(
                    OutOfDomainError,
                    match=(
                        "Invalid inner DataFrame: Found invalid value in column"
                        " 'C': Column contains null values."
                    ),
                ),
                {
                    "domain": SparkGroupedDataFrameDomain(
                        _schema_without_nulls, _base_groupby_columns
                    ),
                    "value": {
                        "dataframe": {
                            "data": [(1, "W", 10), (2, "X", 12), (3, "Y", None)],
                            "schema": ["A", "B", "C"],
                        },
                        "group_keys": _base_group_key_args,
                    },
                },
            ),
            (  # Unexpected nulls in group_keys
                SparkGroupedDataFrameDomain(
                    _schema_without_nulls, _base_groupby_columns
                ),
                {
                    "dataframe": {
                        "data": [(1, "W", 10), (2, "X", 12), (3, "Y", 13)],
                        "schema": ["A", "B", "C"],
                    },
                    "group_keys": _group_key_with_nulls_args,
                },
                pytest.raises(
                    OutOfDomainError,
                    match=(
                        "Invalid group keys: Found invalid value in column 'B': "
                        "Column contains null values."
                    ),
                ),
                {
                    "domain": SparkGroupedDataFrameDomain(
                        _schema_without_nulls, _base_groupby_columns
                    ),
                    "value": {
                        "dataframe": {
                            "data": [(1, "W", 10), (2, "X", 12), (3, "Y", 13)],
                            "schema": ["A", "B", "C"],
                        },
                        "group_keys": _group_key_with_nulls_args,
                    },
                },
            ),
            (  # Missing column in dataframe
                SparkGroupedDataFrameDomain(_base_schema, _base_groupby_columns),
                {
                    "dataframe": {
                        "data": [(1, "W"), (2, "X"), (3, "Y")],
                        "schema": ["A", "B"],
                    },
                    "group_keys": _base_group_key_args,
                },
                pytest.raises(
                    OutOfDomainError,
                    match=(
                        "Invalid inner DataFrame: Columns are not as expected. "
                        "DataFrame and Domain must contain the same columns in the "
                        "same order.\nDataFrame columns: \\['A', 'B'\\]\nDomain "
                        "columns: \\['A', 'B', 'C'\\]"
                    ),
                ),
                {
                    "domain": SparkGroupedDataFrameDomain(
                        _base_schema, _base_groupby_columns
                    ),
                    "value": {
                        "dataframe": {
                            "data": [(1, "W"), (2, "X"), (3, "Y")],
                            "schema": ["A", "B"],
                        },
                        "group_keys": _base_group_key_args,
                    },
                },
            ),
            (  # Missing column in group_keys
                SparkGroupedDataFrameDomain(_base_schema, _base_groupby_columns),
                {
                    "dataframe": {
                        "data": [(1, "W", 10), (2, "X", 12), (3, "Y", 13)],
                        "schema": ["A", "B", "C"],
                    },
                    "group_keys": _empty_group_key_args,
                },
                pytest.raises(
                    OutOfDomainError,
                    match=(
                        "Invalid group keys: Columns are not as expected. "
                        "DataFrame and Domain must contain the same columns in the "
                        "same order.\nDataFrame columns: \\[\\]\nDomain "
                        "columns: \\['A', 'B'\\]"
                    ),
                ),
                {
                    "domain": SparkGroupedDataFrameDomain(
                        _base_schema, _base_groupby_columns
                    ),
                    "value": {
                        "dataframe": {
                            "data": [(1, "W", 10), (2, "X", 12), (3, "Y", 13)],
                            "schema": ["A", "B", "C"],
                        },
                        "group_keys": _empty_group_key_args,
                    },
                },
            ),
            (  # empty group_keys
                SparkGroupedDataFrameDomain(_base_schema, []),
                {
                    "dataframe": {
                        "data": [(1, "W", 10), (2, "X", 12), (3, "Y", 13)],
                        "schema": ["A", "B", "C"],
                    },
                    "group_keys": _empty_group_key_args,
                },
                does_not_raise(),
                None,
            ),
        ],
    )
    def test_validate(
        self,
        domain: Domain,
        candidate: Any,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Validate works correctly.

        Args:
            domain: The domain to test.
            candidate: The value to validate using domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        candidate["dataframe"] = self.spark.createDataFrame(**candidate["dataframe"])
        candidate["group_keys"] = self.spark.createDataFrame(**candidate["group_keys"])
        candidate = GroupedDataFrame(**candidate)
        if exception_properties is not None:
            exception_properties["value"]["dataframe"] = self.spark.createDataFrame(
                **exception_properties["value"]["dataframe"]
            )
            exception_properties["value"]["group_keys"] = self.spark.createDataFrame(
                **exception_properties["value"]["group_keys"]
            )
            exception_properties["value"] = GroupedDataFrame(
                **exception_properties["value"]
            )
        with expectation as exception:
            domain.validate(candidate)
        if exception_properties is None or len(exception_properties) == 0:
            return
        # Help out mypy
        assert isinstance(exception, pytest.ExceptionInfo)
        core_exception = exception.value
        assert hasattr(core_exception, "domain"), "Exception has no domain attribute"
        assert (
            exception_properties["domain"] == core_exception.domain
        ), "Exception domain is not as expected"
        assert hasattr(core_exception, "value"), "Exception has no value attribute"
        # Separate asserts for the frames are required since GroupedDataFrame does not
        # implement __eq__.
        assert_frame_equal_with_sort(
            exception_properties[  # pylint: disable=protected-access
                "value"
            ]._dataframe,
            core_exception.value._dataframe,  # pylint: disable=protected-access
        )
        assert_frame_equal_with_sort(
            exception_properties["value"].group_keys, core_exception.value.group_keys
        )

    @pytest.mark.parametrize(
        "domain", [SparkGroupedDataFrameDomain(_base_schema, _base_groupby_columns)]
    )
    def test_repr(self, domain: Domain):  # pylint: disable=no-self-use
        """Tests that __repr__ works correctly."""
        expected = (
            "SparkGroupedDataFrameDomain(schema={'A': SparkIntegerColumnDescriptor("
            "allow_null=True, size=64), 'B': SparkStringColumnDescriptor(allow_null="
            "True), 'C': SparkIntegerColumnDescriptor(allow_null=True, size=64)},"
            " groupby_columns=['A', 'B'])"
        )
        assert repr(domain) == expected

    @pytest.mark.parametrize(
        "domain, expected, expectation, exception_properties",
        [
            (
                SparkGroupedDataFrameDomain(
                    schema={
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                        "C": SparkFloatColumnDescriptor(),
                    },
                    groupby_columns=["A", "B"],
                ),
                SparkDataFrameDomain(schema={"C": SparkFloatColumnDescriptor()}),
                does_not_raise(),
                None,
            )
        ],
    )
    def test_get_group_domain(  # pylint: disable=no-self-use
        self,
        domain: SparkGroupedDataFrameDomain,
        expected: SparkDataFrameDomain,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """get_group_domain returns the correct group domain."""
        with expectation as exception:
            assert domain.get_group_domain() == expected
        if exception_properties is None:
            return
        for key, value in exception_properties.items():
            assert hasattr(exception, key)
            assert getattr(exception, key) == value


class TestSparkRowDomain(DomainTests):
    """Tests for :class:`~tmlt.core.domains.spark_domains.SparkRowDomain`."""

    @pytest.fixture
    def domain_type(self) -> Type[Domain]:  # pylint: disable=no-self-use
        """Returns the type of the domain to be tested."""
        return SparkRowDomain

    @pytest.fixture
    def domain(self) -> SparkRowDomain:  # pylint: disable=no-self-use
        """Get a base SparkRowDomain."""
        return SparkRowDomain(
            schema={
                "A": SparkIntegerColumnDescriptor(),
                "B": SparkStringColumnDescriptor(),
            }
        )

    @pytest.mark.parametrize(
        "domain_args, expectation, exception_properties",
        [
            (
                {
                    "schema": {
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                },
                does_not_raise(),
                None,
            ),
            (
                {"schema": int},
                pytest.raises(
                    TypeError,
                    match=f'type of argument "schema" must be {get_fullname(Mapping)}; '
                    f"got {get_fullname(int)} instead",
                ),
                None,
            ),
            (
                {"schema": StringType},
                pytest.raises(
                    TypeError,
                    match=f'type of argument "schema" must be {get_fullname(Mapping)}; '
                    f"got {get_fullname(StringType)} instead",
                ),
                None,
            ),
        ],
    )
    def test_construct_component(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Initialization behaves correctly.

        The domain is constructed correctly and raises exceptions when initialized with
        invalid inputs.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        super().test_construct_component(
            domain_type, domain_args, expectation, exception_properties
        )

    @pytest.mark.parametrize(
        "domain, other_domain, expected",
        [
            (
                SparkRowDomain(
                    schema={
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                ),
                SparkRowDomain(
                    schema={
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                ),
                True,
            ),
            (
                # testing that order does matter
                SparkRowDomain(
                    schema={
                        "B": SparkStringColumnDescriptor(),
                        "A": SparkIntegerColumnDescriptor(),
                    }
                ),
                SparkRowDomain(
                    schema={
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                ),
                False,
            ),
            (
                SparkRowDomain(schema={"A": SparkIntegerColumnDescriptor()}),
                SparkRowDomain(schema={"B": SparkStringColumnDescriptor()}),
                False,
            ),
        ],
    )
    def test_eq(self, domain: Domain, other_domain: Domain, expected: bool):
        """__eq__ works correctly.

        Args:
            domain: The domain to test.
            other_domain: The domain to compare to.
            expected: The expected result of the comparison.
        """
        super().test_eq(domain, other_domain, expected)

    @pytest.mark.parametrize(
        "domain_args, key, mutator",
        [
            (
                {
                    "schema": {
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                },
                "schema",
                lambda x: x.update({"A": SparkFloatColumnDescriptor()}),
            )
        ],
    )
    def test_mutable_inputs(
        self,
        domain_type: Type[Domain],
        domain_args: Dict[str, Any],
        key: str,
        mutator: Callable[[Any], Any],
    ):
        """The mutable inputs to the domain are copied.

        Args:
            domain_type: The type of domain to be constructed.
            domain_args: The arguments to the domain.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        super().test_mutable_inputs(domain_type, domain_args, key, mutator)

    @pytest.mark.parametrize(
        "domain, expected_properties",
        [
            (
                SparkRowDomain(
                    schema={
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                ),
                {
                    "schema": {
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    },
                    "carrier_type": Row,
                },
            )
        ],
    )
    def test_properties(self, domain: Domain, expected_properties: Dict[str, Any]):
        """All properties have the expected values.

        Args:
            domain: The constructed domain to be tested.
            expected_properties: A dictionary containing all the property:value pairs
                domain is expected to have.
        """
        super().test_properties(domain, expected_properties)

    def test_property_immutability(self, domain: Domain):
        """The properties return copies for mutable values.

        Args:
            domain: The domain to be tested.
        """

    @pytest.mark.skip(reason="SparkRowDomain does not implement validate.")
    @pytest.mark.parametrize("domain, candidate, expectation, exception_properties", [])
    def test_validate(
        self,
        domain: Domain,
        candidate: Any,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Validate works correctly.

        Args:
            domain: The domain to test.
            candidate: The value to validate using domain.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        super().test_validate(domain, candidate, expectation, exception_properties)

    def test_repr(self, domain: Domain):  # pylint: disable=no-self-use
        """Tests that __repr__ works correctly."""
        expected = (
            "SparkRowDomain(schema={'A': SparkIntegerColumnDescriptor(allow_null=False,"
            " size=64), 'B': SparkStringColumnDescriptor(allow_null=False)})"
        )
        assert repr(domain) == expected


_column_descriptors = {
    "int32": SparkIntegerColumnDescriptor(size=32),
    "int64": SparkIntegerColumnDescriptor(size=64),
    "float32": SparkFloatColumnDescriptor(size=32),
    "str": SparkStringColumnDescriptor(allow_null=True),
    "date": SparkDateColumnDescriptor(),
    "timestamp": SparkTimestampColumnDescriptor(),
}
_col_name_to_type: Dict[str, str] = {
    "A": "int32",
    "B": "int64",
    "C": "float32",
    "D": "str",
    "E": "date",
    "F": "timestamp",
}
_type_to_spark_type: Dict[str, DataType] = {
    "int32": IntegerType(),
    "int64": LongType(),
    "float32": FloatType(),
    "str": StringType(),
    "date": DateType(),
    "timestamp": TimestampType(),
}


@pytest.mark.usefixtures("class_spark")
class TestSparkColumnDescriptors:
    r"""Tests for subclasses of class SparkColumnDescriptor.

    See subclasses of
    :class:`~tmlt.core.domains.spark_domains.SparkColumnDescriptor`\ s."""

    spark: SparkSession

    @pytest.fixture
    def test_df(self) -> DataFrame:
        """Get a base DataFrame"""
        return self.spark.createDataFrame(
            [
                (
                    1,
                    2,
                    1.0,
                    "X",
                    datetime.date.fromisoformat("1970-01-01"),
                    datetime.datetime.fromisoformat("1970-01-01 00:00:00.000+00:00"),
                ),
                (
                    11,
                    239,
                    2.0,
                    None,
                    datetime.date.fromisoformat("2022-01-01"),
                    datetime.datetime.fromisoformat("2022-01-01 08:30:00.000+00:00"),
                ),
            ],
            schema=StructType(
                [
                    StructField("A", IntegerType(), False),
                    StructField("B", LongType(), False),
                    StructField("C", FloatType(), False),
                    StructField("D", StringType(), True),
                    StructField("E", DateType(), True),
                    StructField("F", TimestampType(), True),
                ]
            ),
        )

    @pytest.mark.parametrize(
        "descriptor, expected_domain",
        [
            (SparkIntegerColumnDescriptor(size=32), NumpyIntegerDomain(size=32)),
            (SparkIntegerColumnDescriptor(size=64), NumpyIntegerDomain(size=64)),
            (SparkFloatColumnDescriptor(size=64), NumpyFloatDomain(size=64)),
            (
                SparkFloatColumnDescriptor(size=64, allow_inf=True),
                NumpyFloatDomain(size=64, allow_inf=True),
            ),
            (
                SparkFloatColumnDescriptor(size=64, allow_nan=True),
                NumpyFloatDomain(size=64, allow_nan=True),
            ),
            (
                SparkStringColumnDescriptor(allow_null=True),
                NumpyStringDomain(allow_null=True),
            ),
            (
                SparkStringColumnDescriptor(allow_null=False),
                NumpyStringDomain(allow_null=False),
            ),
        ],
    )
    def test_to_numpy_domain(  # pylint: disable=no-self-use
        self, descriptor: SparkColumnDescriptor, expected_domain: Domain
    ):
        """Tests that to_numpy_domain works correctly."""
        assert descriptor.to_numpy_domain() == expected_domain

    @pytest.mark.parametrize(
        "descriptor, expectation",
        [
            (
                SparkIntegerColumnDescriptor(allow_null=True),
                pytest.raises(
                    RuntimeError,
                    match="Nullable column does not have corresponding NumPy domain.",
                ),
            ),
            (
                SparkDateColumnDescriptor(),
                pytest.raises(
                    RuntimeError, match="NumPy does not have support for date types."
                ),
            ),
            (
                SparkTimestampColumnDescriptor(),
                pytest.raises(
                    RuntimeError,
                    match="NumPy does not have support for timestamp types.",
                ),
            ),
        ],
    )
    def test_to_numpy_domain_invalid(  # pylint: disable=no-self-use
        self, descriptor: SparkColumnDescriptor, expectation: ContextManager[None]
    ):
        """Tests that to_numpy_domain raises appropriate exceptions."""
        with expectation:
            descriptor.to_numpy_domain()

    @pytest.mark.parametrize(
        "descriptor, col_name, expectation",
        [
            (
                _column_descriptors[col_type],
                col_name,
                does_not_raise()
                if col_type == _col_name_to_type[col_name]
                else pytest.raises(
                    ValueError,
                    match="Column must be "
                    f"{get_fullname(_type_to_spark_type[col_type])}; got "
                    f"{get_fullname(_type_to_spark_type[_col_name_to_type[col_name]])} "
                    "instead",
                ),
            )
            for col_name, col_type in product(
                _col_name_to_type.keys(), _col_name_to_type.values()
            )
        ],
    )
    def test_validate_column(  # pylint: disable=no-self-use
        self,
        test_df: DataFrame,
        descriptor: SparkColumnDescriptor,
        col_name: str,
        expectation: ContextManager[None],
    ):
        """Tests that validate_column works correctly."""
        with expectation:
            descriptor.validate_column(test_df, col_name)

    @pytest.mark.parametrize(
        "domain, other_domain, expected",
        [
            (
                _column_descriptors[base_key],
                _column_descriptors[other_key],
                base_key == other_key,
            )
            for base_key, other_key in combinations_with_replacement(
                _col_name_to_type.values(), 2
            )
        ]
        + [
            (
                SparkIntegerColumnDescriptor(size=32),
                SparkIntegerColumnDescriptor(size=32, allow_null=True),
                False,
            )
        ],
    )
    def test_eq(  # pylint: disable=no-self-use
        self,
        domain: SparkColumnDescriptor,
        other_domain: SparkColumnDescriptor,
        expected: bool,
    ):
        """Tests that __eq__ works correctly."""
        assert (domain == other_domain) == expected


class TestSparkUtilityFunctions:
    """Tests for utility functions for creating Spark domains."""

    @pytest.mark.parametrize(
        "spark_schema, expected, expectation",
        [
            (StructType(), {}, does_not_raise()),
            (
                StructType([StructField("A", IntegerType(), False)]),
                {"A": SparkIntegerColumnDescriptor(size=32)},
                does_not_raise(),
            ),
            (
                StructType(
                    [
                        StructField("A", IntegerType(), False),
                        StructField("B", LongType(), False),
                        StructField("C", FloatType(), False),
                        StructField("D", StringType(), True),
                        StructField("E", DateType(), True),
                        StructField("F", TimestampType(), True),
                    ]
                ),
                {
                    "A": SparkIntegerColumnDescriptor(size=32),
                    "B": SparkIntegerColumnDescriptor(size=64),
                    "C": SparkFloatColumnDescriptor(
                        allow_nan=True, allow_inf=True, size=32
                    ),
                    "D": SparkStringColumnDescriptor(allow_null=True),
                    "E": SparkDateColumnDescriptor(allow_null=True),
                    "F": SparkTimestampColumnDescriptor(allow_null=True),
                },
                does_not_raise(),
            ),
        ],
    )
    def test_convert_spark_schema(  # pylint: disable=no-self-use
        self,
        spark_schema: StructType,
        expected: SparkColumnsDescriptor,
        expectation: ContextManager[None],
    ):
        """Tests that convert_spark_schema works correctly.

        Args:
            spark_schema: The Spark schema to convert.
            expected: The expected result of the conversion.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        with expectation:
            assert convert_spark_schema(spark_schema) == expected

    @pytest.mark.parametrize(
        "pandas_domain, expected, expectation",
        [
            (PandasDataFrameDomain({}), {}, does_not_raise()),
            (
                PandasDataFrameDomain(
                    {
                        "A": PandasSeriesDomain(NumpyIntegerDomain(size=32)),
                        "B": PandasSeriesDomain(NumpyIntegerDomain(size=64)),
                        "C": PandasSeriesDomain(
                            NumpyFloatDomain(allow_nan=True, allow_inf=True, size=32)
                        ),
                        "D": PandasSeriesDomain(NumpyStringDomain()),
                    }
                ),
                {
                    "A": SparkIntegerColumnDescriptor(size=32),
                    "B": SparkIntegerColumnDescriptor(size=64),
                    "C": SparkFloatColumnDescriptor(
                        allow_nan=True, allow_inf=True, size=32
                    ),
                    "D": SparkStringColumnDescriptor(),
                },
                does_not_raise(),
            ),
        ],
    )
    def test_convert_pandas_domain(  # pylint: disable=no-self-use
        self,
        pandas_domain: PandasDataFrameDomain,
        expected: SparkColumnsDescriptor,
        expectation: ContextManager[None],
    ):
        """Tests that convert_pandas_domain works correctly.

        Args:
            pandas_domain: The Pandas domain to convert.
            expected: The expected result of the conversion.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        with expectation:
            assert convert_pandas_domain(pandas_domain) == expected

    @pytest.mark.parametrize(
        "numpy_domain, expected, expectation",
        [
            (
                NumpyIntegerDomain(size=32),
                SparkIntegerColumnDescriptor(size=32),
                does_not_raise(),
            ),
            (
                NumpyFloatDomain(),
                SparkFloatColumnDescriptor(allow_nan=False, allow_inf=False, size=64),
                does_not_raise(),
            ),
            (
                NumpyFloatDomain(allow_nan=True, allow_inf=True, size=32),
                SparkFloatColumnDescriptor(allow_nan=True, allow_inf=True, size=32),
                does_not_raise(),
            ),
            (NumpyStringDomain(), SparkStringColumnDescriptor(), does_not_raise()),
            (
                NumpyStringDomain(allow_null=True),
                SparkStringColumnDescriptor(allow_null=True),
                does_not_raise(),
            ),
            ("Not a numpy domain", None, pytest.raises(NotImplementedError)),
        ],
    )
    def test_convert_numpy_domain(  # pylint: disable=no-self-use
        self,
        numpy_domain: NumpyDomain,
        expected: SparkColumnDescriptor,
        expectation: ContextManager[None],
    ):
        """Tests that convert_numpy_domain works correctly.

        Args:
            numpy_domain: The Numpy domain to convert.
            expected: The expected result of the conversion.
            expectation: A context manager that captures the correct expected type of
                error that is raised.
        """
        with expectation:
            assert convert_numpy_domain(numpy_domain) == expected



"""Wrappers for changing a transformation's output metric."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from typing import Any

from pyspark.sql import DataFrame
from typeguard import typechecked

from tmlt.core.domains.spark_domains import SparkDataFrameDomain
from tmlt.core.exceptions import DomainColumnError, UnsupportedMetricError
from tmlt.core.metrics import (
    HammingDistance,
    IfGroupedBy,
    RootSumOfSquared,
    SumOf,
    SymmetricDifference,
)
from tmlt.core.transformations.base import Transformation
from tmlt.core.utils.exact_number import ExactNumber, ExactNumberInput


class UnwrapIfGroupedBy(Transformation):
    """No-op transformation for switching from IfGroupedBy to its inner metric."""

    @typechecked
    def __init__(self, domain: SparkDataFrameDomain, input_metric: IfGroupedBy):
        """Constructor.

        Args:
            domain: Domain of input DataFrames.
            input_metric: IfGroupedBy metric on input DataFrames.
        """
        if not input_metric.column in domain.schema:
            raise DomainColumnError(
                domain,
                input_metric.column,
                f"Invalid IfGroupedBy metric: {input_metric.column} not in domain",
            )
        if not isinstance(input_metric.inner_metric, (SumOf, RootSumOfSquared)):
            raise UnsupportedMetricError(
                input_metric,
                (
                    "Inner metric for IfGroupedBy metric must be "
                    "SumOf(SymmetricDifference()), or "
                    "RootSumOfSquared(SymmetricDifference())"
                ),
            )
        self._is_l2 = isinstance(input_metric.inner_metric, RootSumOfSquared)
        super().__init__(
            input_domain=domain,
            input_metric=input_metric,
            output_domain=domain,
            output_metric=input_metric.inner_metric.inner_metric,
        )

    @typechecked
    def stability_function(self, d_in: ExactNumberInput) -> ExactNumber:
        r"""Returns the smallest d_out satisfied by the transformation.

        If the inner metric of the :class:`~.IfGroupedBy` input metric is a
        :class:`~.SumOf`, returns `d_in`.

        If the inner metric is :class:`~.RootSumOfSquared`, returns `d_in`\*\*2.

        Args:
            d_in: Distance between inputs under input_metric.
        """
        self.input_metric.validate(d_in)
        d_in = ExactNumber(d_in)
        if self._is_l2:
            return d_in**2
        return d_in

    def __call__(self, sdf: DataFrame) -> DataFrame:
        """Returns DataFrame unchanged."""
        return sdf


class HammingDistanceToSymmetricDifference(Transformation):
    """No-op transformation for switching metrics."""

    @typechecked
    def __init__(self, domain: SparkDataFrameDomain):
        """Constructor."""
        super().__init__(
            input_domain=domain,
            input_metric=HammingDistance(),
            output_domain=domain,
            output_metric=SymmetricDifference(),
        )

    @typechecked
    def stability_function(self, d_in: ExactNumberInput) -> ExactNumber:
        """Returns the smallest d_out satisfied by the transformation.

        Returns 2 * d_in

        Args:
            d_in: Distance between inputs under input_metric.
        """
        self.input_metric.validate(d_in)
        return ExactNumber(2) * d_in

    def __call__(self, data: Any):
        """Returns unchanged input."""
        return data



"""Tumult Core Module."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023
import warnings

from tmlt.core.utils.configuration import check_java11

warnings.filterwarnings(action="ignore", category=UserWarning, message=".*open_stream")
warnings.filterwarnings(
    action="ignore", category=FutureWarning, message=".*check_less_precise.*"
)

check_java11()



"""Unit tests for :mod:`tmlt.core.utils.configuration`."""

import os
import re
import subprocess
from string import ascii_letters, digits
from typing import Any, Optional
from unittest import TestCase
from unittest.mock import ANY, MagicMock, call, patch

from parameterized import parameterized
from pyspark.sql import SparkSession

from tmlt.core.utils.configuration import (
    Config,
    SparkConfigError,
    _get_java_version,
    _java11_config_opts,
    _simple_java_version,
    check_java11,
    get_java11_config,
)

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023


class TestConfiguration(TestCase):
    """TestCase for Config."""

    def test_db_name(self):
        """Config.temp_db_name() returns a valid db name."""
        self.assertIsInstance(Config.temp_db_name(), str)
        self.assertTrue(len(Config.temp_db_name()) > 0)
        self.assertIn(Config.temp_db_name()[0], ascii_letters + digits)

    # pylint: disable=protected-access
    @parameterized.expand(
        [
            (
                (
                    'openjdk version "1.8.0_292"\nOpenJDK Runtime Environment'
                    " (AdoptOpenJDK)(build 1.8.0_292-b10)\nOpenJDK 64-Bit Server VM"
                    " (AdoptOpenJDK)(build 25.292-b10, mixed mode)"
                ),
                8,
            ),
            ("1.2.345_678", 2),
            ('un-open jdk version "11.987.654_321"', 11),
            ("15.43.21", 15),
            ("totally bogus version string", None),
        ]
    )
    def test_simple_java_version(
        self, version_str: str, expected: Optional[int]
    ) -> None:
        """Test _simple_java_version."""
        if expected is None:
            with self.assertWarnsRegex(
                RuntimeWarning,
                re.escape(
                    "Unable to determine Java version from version string"
                    f" `{version_str}`"
                ),
            ):
                self.assertEqual(_simple_java_version(version_str), expected)
        else:
            self.assertEqual(_simple_java_version(version_str), expected)

    @patch("tmlt.core.utils.configuration.subprocess.check_output")
    def test_subprocess_path_overwritten(self, mock_check_output) -> None:
        """Test that when JAVA_HOME is set, JAVA_HOME/bin is prepended to PATH."""
        self.assertIsNone(SparkSession.getActiveSession())

        # Check that nothing changes when JAVA_HOME is unset
        orig_java_home = os.environ.get("JAVA_HOME")
        if orig_java_home:
            del os.environ["JAVA_HOME"]
        environ = os.environ.copy()
        mock_check_output.return_value = '"1.2.345"'
        _get_java_version()
        mock_check_output.assert_called_once()
        mock_check_output.assert_called_with(
            ["java", "-version"], stderr=subprocess.STDOUT, env=environ
        )

        # Check that nothing changes when JAVA_HOME is the empty string
        mock_check_output.reset_mock()
        os.environ["JAVA_HOME"] = ""
        environ = os.environ.copy()
        _get_java_version()
        mock_check_output.assert_called_once()
        mock_check_output.assert_called_with(
            ["java", "-version"], stderr=subprocess.STDOUT, env=environ
        )

        # Check that when JAVA_HOME is set, subprocess gets an environment
        # where the PATH starts with JAVA_HOME/bin
        mock_check_output.reset_mock()
        java_home = os.path.join(os.getcwd(), "fake_java_home")
        os.environ["JAVA_HOME"] = java_home
        environ = os.environ.copy()
        orig_path = os.environ.get("PATH")
        expected_new_path = f"{os.path.join(java_home, 'bin')}:{orig_path}"
        environ["PATH"] = expected_new_path
        _get_java_version()
        mock_check_output.assert_called_once()
        mock_check_output.assert_called_with(
            ["java", "-version"], stderr=subprocess.STDOUT, env=environ
        )

        # Reset environment (if JAVA_HOME was set at the beginning of this test)
        if orig_java_home:
            os.environ["JAVA_HOME"] = orig_java_home

    @parameterized.expand(
        [
            ("1.8.0_292", 8),
            ("1.2.345_678", 2),
            ("11.987.654_321", 11),
            ("15.43.21", 15),
            ("totally bogus version string", None),
        ]
    )
    @patch("tmlt.core.utils.configuration.SparkSession.getActiveSession")
    def test_get_java_version_from_spark(
        self, str_from_spark: str, expected: Optional[int], mock_get_active_session
    ) -> None:
        """Test _get_java_version when Spark is running."""
        # pylint: disable=line-too-long
        mock_get_active_session.return_value.sparkContext._jvm.System.getProperty.return_value = (
            str_from_spark
        )
        # pylint: enable=line-too-long
        if expected is None:
            with self.assertWarnsRegex(
                RuntimeWarning, "Unable to determine Java version from version string"
            ):
                self.assertEqual(_get_java_version(), expected)
        else:
            self.assertEqual(_get_java_version(), expected)

    @parameterized.expand(
        [
            (
                (
                    'openjdk version "1.8.0_292"\nOpenJDK Runtime Environment'
                    " (AdoptOpenJDK)(build 1.8.0_292-b10)\nOpenJDK 64-Bit Server VM"
                    " (AdoptOpenJDK)(build 25.292-b10, mixed mode)"
                ),
                8,
            ),
            ('Artisanal Handmade Java "1.2.345_678"', 2),
            ('un-open jdk version "11.987.654_321"', 11),
            ('some other kind of java "15.43.21"', 15),
            ("totally bogus version string", None),
        ]
    )
    @patch("tmlt.core.utils.configuration.subprocess.check_output")
    def test_get_java_version_from_subprocess(
        self,
        str_from_subprocess: str,
        expected: Optional[int],
        mock_subprocess_check_output,
    ) -> None:
        """Test _get_java_version when Spark is not running."""
        self.assertIsNone(SparkSession.getActiveSession())
        mock_subprocess_check_output.return_value = str_from_subprocess
        if expected is None:
            with self.assertWarnsRegex(
                RuntimeWarning, "Unable to determine Java version from version string"
            ):
                self.assertEqual(_get_java_version(), expected)
        else:
            self.assertEqual(_get_java_version(), expected)

    def test_get_java11_config(self) -> None:
        """Test that the java11 config has all the java11 options."""
        java11_config = get_java11_config()
        for k, v in _java11_config_opts().items():
            self.assertEqual(java11_config.get(k), v)

    @patch("tmlt.core.utils.configuration._get_java_version")
    @patch("tmlt.core.utils.configuration.SparkSession.builder")
    @patch("tmlt.core.utils.configuration.SparkSession.getActiveSession")
    def test_check_java11(
        self, mock_get_active_session, mock_builder, mock_get_java_version
    ) -> None:
        """Test the Java 11 checker."""

        # First, make sure nothing happens if the Java version is less than 11
        mock_get_java_version.return_value = 0
        mock_get_active_session.return_value = None
        check_java11()
        mock_get_active_session.assert_not_called()
        mock_builder.config.assert_not_called()

        # reset mocks
        mock_get_active_session.reset_mock()
        mock_builder.reset_mock()

        # Now, check what happens if the version is >= 11 and Spark is not running
        mock_get_java_version.return_value = 11
        mock_get_active_session.return_value = None
        check_java11()
        self.assertEqual(mock_get_active_session.call_count, 1)
        self.assertEqual(mock_builder.config.call_count, len(_java11_config_opts()))
        expected_builder_config_calls = [
            call(k, v) for k, v in _java11_config_opts().items()
        ]
        mock_builder.config.assert_has_calls(
            expected_builder_config_calls, any_order=True
        )

        # reset mocks
        mock_get_active_session.reset_mock()
        mock_builder.reset_mock()

        # Now, check what happens when the config values are set correctly
        mock_get_java_version.return_value = 11
        mock_get_active_session.return_value = MagicMock(spec=SparkSession)

        def mock_config_get(key: str, default: Any) -> Any:
            return _java11_config_opts().get(key, default)

        mock_get_active_session.return_value.conf.get = MagicMock(
            side_effect=mock_config_get
        )
        expected_conf_get_calls = [call(k, ANY) for k in _java11_config_opts()]
        check_java11()
        self.assertEqual(mock_get_active_session.call_count, 1)
        self.assertEqual(mock_builder.call_count, 0)
        self.assertEqual(
            mock_get_active_session.return_value.conf.get.call_count,
            len(_java11_config_opts()),
        )
        mock_get_active_session.return_value.conf.get.assert_has_calls(
            expected_conf_get_calls, any_order=True
        )

        # reset mocks
        mock_get_active_session.reset_mock()
        mock_builder.reset_mock()

        # If Spark has started and the values aren't set, there should be an error
        mock_get_java_version.return_value = 11
        mock_get_active_session.return_value = MagicMock(spec=SparkSession)
        mock_get_active_session.return_value.conf.get.return_value = "bad config opt"
        with self.assertRaisesRegex(
            SparkConfigError,
            re.escape(
                "When running Spark on Java 11 or higher, you need to set up your Spark"
                " session with specific configuration options *before* you start Spark."
                " Core automatically sets these options on `SparkSession.builder` if"
                " you import Core before you build your session."
            ),
        ):
            check_java11()



"""Domains for NumPy datatypes."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from dataclasses import dataclass
from typing import Any

import numpy as np
from typeguard import check_type

from tmlt.core.domains.base import Domain, OutOfDomainError


class NumpyDomain(Domain):
    """Base class for NumpyDomains."""

    @classmethod
    def from_np_type(cls, dtype: np.dtype) -> "NumpyDomain":
        """Returns a NumPy domain from a Pandas type."""
        pd_to_numpy_domain = {
            np.dtype("int8"): NumpyIntegerDomain(size=32),
            np.dtype("int16"): NumpyIntegerDomain(size=32),
            np.dtype("int32"): NumpyIntegerDomain(size=32),
            np.dtype("int64"): NumpyIntegerDomain(size=64),
            np.dtype("float32"): NumpyFloatDomain(size=32),
            np.dtype("float64"): NumpyFloatDomain(size=64),
            np.dtype("object"): NumpyStringDomain(),
            np.dtype("bool"): NumpyIntegerDomain(size=32),
        }
        return pd_to_numpy_domain[dtype]


@dataclass(frozen=True)
class NumpyIntegerDomain(NumpyDomain):
    """Domain of NumPy integers."""

    size: int = 64
    """Number of bits a member of the domain occupies. Must be 32 or 64."""

    def __post_init__(self):
        """Checks arguments to constructor."""
        check_type("size", self.size, int)
        if self.size not in [32, 64]:
            raise ValueError(f"size must be 32 or 64, not {self.size}")

    @property
    def carrier_type(self) -> type:
        """Returns carrier type for elements in the domain."""
        if self.size == 32:
            return np.int32
        return np.int64


@dataclass(frozen=True)
class NumpyFloatDomain(NumpyDomain):
    """Domain of NumPy floats."""

    allow_nan: bool = False
    """If True, NaNs are permitted in the domain."""
    allow_inf: bool = False
    """If True, infs are permitted in the domain."""
    size: int = 64
    """Number of bits a member of the domain occupies. Must be 32 or 64."""

    def __post_init__(self):
        """Checks arguments to constructor."""
        check_type("allow_nan", self.allow_nan, bool)
        check_type("allow_inf", self.allow_inf, bool)
        check_type("size", self.size, int)
        if self.size not in [32, 64]:
            raise ValueError(f"size must be 32 or 64, not {self.size}")

    @property
    def carrier_type(self) -> type:
        """Returns carrier type for elements in the domain."""
        if self.size == 32:
            return np.float32
        return np.float64

    def validate(self, value: Any):
        """Raises error if value is not a member of the domain."""
        super().validate(value)
        if not self.allow_inf and np.isinf(value):
            raise OutOfDomainError(self, value, "Value is infinite.")
        if not self.allow_nan and np.isnan(value):
            raise OutOfDomainError(self, value, "Value is NaN.")


@dataclass(frozen=True)
class NumpyStringDomain(NumpyDomain):
    """Domain of NumPy strings.

    Note:
        This domain does not use :class:`numpy.str_` datatype. Instead, native python
        objects are used since Pandas uses this by default.
    """

    allow_null: bool = False
    """If True, None is allowed."""

    def __post_init__(self):
        """Checks arguments to constructor."""
        check_type("allow_null", self.allow_null, bool)

    def validate(self, value: Any):
        """Raises error if value is not in domain."""
        if not isinstance(value, self.carrier_type) and value.__class__ is not str:
            raise OutOfDomainError(
                self, value, f"Value must be {str}, instead it is {value.__class__}."
            )
        if value is None and not self.allow_null:
            raise OutOfDomainError(self, value, "Value is null.")

    @property
    def carrier_type(self) -> type:  # pylint: disable=no-self-use
        """Returns carrier types for members of NumpyStringDomain."""
        return object



"""Domains for Pandas datatypes."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from collections import OrderedDict
from dataclasses import dataclass
from typing import Any, Dict

import numpy as np
import pandas as pd
from typeguard import check_type, typechecked

from tmlt.core.domains.base import Domain, OutOfDomainError
from tmlt.core.domains.numpy_domains import NumpyDomain


@dataclass(frozen=True)
class PandasSeriesDomain(Domain):
    """Domain of Pandas Series.

    Note:
        The index is always ignored when this domain type is used.
    """

    element_domain: NumpyDomain
    """Domain of elements in the Series."""

    def __post_init__(self):
        """Checks arguments to constructor."""
        check_type("element_domain", self.element_domain, NumpyDomain)

    @property
    def carrier_type(self) -> type:
        """Returns carrier type for members of the domain."""
        return pd.Series

    def validate(self, value: Any):
        """Raises error if value is not a DataFrame with matching schema."""
        # NOTE: Can not assert (elem in self.element_domain for elem in value) because
        # iterating over a Series implicitly calls item() on the NumPy values
        # retrieving the corresponding python object
        super().validate(value)
        for i in range(len(value)):  # pylint: disable=consider-using-enumerate
            try:
                self.element_domain.validate(value[i])
            except OutOfDomainError as exception:
                raise OutOfDomainError(
                    self, value, f"Found invalid value in Series: {exception}"
                ) from exception

    @classmethod
    def from_numpy_type(cls, dtype: np.dtype) -> "PandasSeriesDomain":
        """Returns a Pandas Series from a NumPy type."""
        return PandasSeriesDomain(NumpyDomain.from_np_type(dtype))


PandasColumnsDescriptor = Dict[str, PandasSeriesDomain]
"""Mapping from column name to column domain."""


class PandasDataFrameDomain(Domain):
    """Domain of Pandas DataFrames."""

    @typechecked
    def __init__(self, schema: PandasColumnsDescriptor):
        """Constructor.

        Args:
            schema: Mapping from column name to column domain.
        """
        self._schema = schema.copy()

    def __repr__(self) -> str:
        """Return string representation of the object."""
        return f"{self.__class__.__name__}(schema={self._schema})"

    @property
    def schema(self) -> PandasColumnsDescriptor:
        """Returns mapping from column name to associated domain."""
        return self._schema.copy()

    @property
    def carrier_type(self) -> type:
        """Returns carrier type for the domain."""
        return pd.DataFrame

    def validate(self, value: Any):
        """Raises error if value is not a Pandas DataFrame with matching schema."""
        super().validate(value)
        value_columns = list(value.columns)
        if len(value_columns) > len(set(value_columns)):
            duplicates = set(
                col for col in value_columns if value_columns.count(col) > 1
            )
            raise OutOfDomainError(
                self, value, f"Some columns are duplicated, {sorted(duplicates)}"
            )

        schema_columns = list(self.schema.keys())
        if value_columns != schema_columns:
            raise OutOfDomainError(
                self,
                value,
                (
                    "Columns are not as expected. DataFrame and Domain must contain"
                    " the same columns in the same order.\nDataFrame columns:"
                    f" {value_columns}\nDomain columns: {schema_columns}"
                ),
            )

        for column in self.schema:
            try:
                self.schema[column].validate(value[column])
            except OutOfDomainError as exception:
                raise OutOfDomainError(
                    self,
                    value,
                    f"Found invalid value in column '{column}': {exception}",
                ) from exception

    def __eq__(self, other: Any) -> bool:
        """Return True if the classes are equivalent."""
        if self.__class__ != other.__class__:
            return False
        return OrderedDict(self.schema) == OrderedDict(other.schema)

    @classmethod
    def from_numpy_types(cls, dtypes: Dict[str, np.dtype]) -> "PandasDataFrameDomain":
        """Returns a Pandas DataFrame domain from a dictionary of NumPy types."""
        col_to_desc = {}
        for col in dtypes:
            col_to_desc[col] = PandasSeriesDomain.from_numpy_type(dtypes[col])
        return PandasDataFrameDomain(col_to_desc)



"""Utilities for checking the inputs to components."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from __future__ import annotations

import datetime
from typing import List, Mapping, Optional, Union

from tmlt.core.domains.spark_domains import SparkDataFrameDomain
from tmlt.core.utils.exact_number import ExactNumber, ExactNumberInput


# You might want to put a type alias here for domains.
# Then you will discover that Sphinx AutoAPI doesn't support type aliases
# the way you want it to.
# Don't fall down this rabbit hole.
def validate_groupby_domains(
    groupby_domains: Mapping[
        str,
        Union[
            List[str],
            List[Optional[str]],
            List[int],
            List[Optional[int]],
            List[datetime.date],
            List[Optional[datetime.date]],
        ],
    ],
    input_domain: SparkDataFrameDomain,
):
    """Raises error if groupby domains are invalid.

    In particular, this passes only if:

        - each column has a non-empty domain AND
        - for each column, all values in its domain are valid values w.r.t the
          input domain AND
        - for each column, each value in its domain is distinct.
    """
    for column, domain in groupby_domains.items():
        if not domain:
            raise ValueError(f"Domain for '{column}' is empty!")
        col_desc = input_domain.schema[column]
        unique_keys = set()
        for key in domain:
            if key in unique_keys:
                raise ValueError(f"Domain for '{column}' contains duplicates.")
            unique_keys.add(key)
            if not col_desc.valid_py_value(key):
                raise ValueError(
                    f"Groupby key '{key}' is invalid for column's '{column}' "
                    f"domain {col_desc}."
                )


def validate_exact_number(
    value: ExactNumberInput,
    allow_nonintegral: bool = True,
    minimum: Optional[ExactNumberInput] = None,
    minimum_is_inclusive: bool = True,
    maximum: Optional[ExactNumberInput] = None,
    maximum_is_inclusive: bool = True,
) -> None:
    """Raises a :class:`ValueError` if `value` fails the specified conditions.

    Examples:
        Verify that a number is integral

        ..
            >>> from fractions import Fraction
            >>> import sympy as sp

        >>> validate_exact_number(
        ...     value=1,
        ...     allow_nonintegral=False,
        ... )
        >>> validate_exact_number(
        ...     value=Fraction(1, 2),
        ...     allow_nonintegral=False,
        ... )
        Traceback (most recent call last):
        ValueError: 1/2 is not an integer

        Verify that a number is between 0 and 1 (inclusive)

        >>> validate_exact_number(
        ...     value=1,
        ...     minimum=0,
        ...     maximum=1,
        ... )
        >>> validate_exact_number(
        ...     value=-1,
        ...     minimum=0,
        ...     maximum=1,
        ... )
        Traceback (most recent call last):
        ValueError: -1 is not greater than or equal to 0
        >>> validate_exact_number(
        ...     value=2,
        ...     minimum=0,
        ...     maximum=1,
        ... )
        Traceback (most recent call last):
        ValueError: 2 is not less than or equal to 1

        Verify that a number is a finite integer

        >>> validate_exact_number(
        ...     value=-123,
        ...     allow_nonintegral=False,
        ...     minimum=-float("inf"),
        ...     minimum_is_inclusive=False,
        ...     maximum=float("inf"),
        ...     maximum_is_inclusive=False,
        ... )
        >>> validate_exact_number(
        ...     value="0.5",
        ...     allow_nonintegral=False,
        ...     minimum=-float("inf"),
        ...     minimum_is_inclusive=False,
        ...     maximum=float("inf"),
        ...     maximum_is_inclusive=False,
        ... )
        Traceback (most recent call last):
        ValueError: 0.5 is not an integer
        >>> validate_exact_number(
        ...     value=sp.oo,
        ...     allow_nonintegral=False,
        ...     minimum=-float("inf"),
        ...     minimum_is_inclusive=False,
        ...     maximum=float("inf"),
        ...     maximum_is_inclusive=False,
        ... )
        Traceback (most recent call last):
        ValueError: oo is not strictly less than inf
        >>> validate_exact_number(
        ...     value=-sp.oo,
        ...     allow_nonintegral=False,
        ...     minimum=-float("inf"),
        ...     minimum_is_inclusive=False,
        ...     maximum=float("inf"),
        ...     maximum_is_inclusive=False,
        ... )
        Traceback (most recent call last):
        ValueError: -oo is not strictly greater than -inf

    Args:
        value: A :class:`sympy.Expr` to validate.
        allow_nonintegral: If False, raises an error if `value` is not integral,
            unless it is infinity.
        minimum: An optional lower bound.
        minimum_is_inclusive: If False, `value` is not allowed to be equal to
            `minimum`. Defaults to True.
        maximum: An optional upper bound.
        maximum_is_inclusive: If False, `value` being equal to `maximum` is not
            allowed. Defaults to True.
    """
    exact_value = ExactNumber(value)
    if not (allow_nonintegral or exact_value.is_integer or not exact_value.is_finite):
        raise ValueError(f"{value} is not an integer")
    if minimum is not None:
        exact_minimum = ExactNumber(minimum)
        if minimum_is_inclusive:
            if exact_value < exact_minimum:
                raise ValueError(f"{value} is not greater than or equal to {minimum}")
        else:
            if exact_value <= exact_minimum:
                raise ValueError(f"{value} is not strictly greater than {minimum}")
    if maximum is not None:
        exact_maximum = ExactNumber(maximum)
        if maximum_is_inclusive:
            if exact_value > exact_maximum:
                raise ValueError(f"{value} is not less than or equal to {maximum}")
        else:
            if exact_value >= exact_maximum:
                raise ValueError(f"{value} is not strictly less than {maximum}")



"""Grouped DataFrame aware of group keys when performing aggregations."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

import functools
from functools import reduce
from typing import Any, Callable, Dict, List

import pandas as pd
from pyspark.sql import Column, DataFrame, Row, SparkSession
from pyspark.sql import functions as sf
from pyspark.sql.types import StructType

from tmlt.core.utils.join import join
from tmlt.core.utils.misc import get_nonconflicting_string


class GroupedDataFrame:
    """Grouped DataFrame implementation supporting explicit group keys.

    A GroupedDataFrame object encapsulates the spark DataFrame to be grouped by as well
    as the group keys. The output of an aggregation on a GroupedDataFrame object is
    guaranteed to have exactly one row for each group key, unless there are no group
    keys, in which case it will have a single row.
    """

    def __init__(self, dataframe: DataFrame, group_keys: DataFrame):
        """Constructor.

        Args:
            dataframe: DataFrame to perform groupby on.
            group_keys: DataFrame where each row corresponds to a group key. Duplicate
                rows are silently dropped.
        """
        if len(dataframe.columns) != len(set(dataframe.columns)):
            raise ValueError("DataFrame contains duplicate column names")
        if len(group_keys.columns) != len(set(group_keys.columns)):
            raise ValueError("Group keys contains duplicate column names")
        invalid_groupby_columns = set(group_keys.columns) - set(dataframe.columns)
        if invalid_groupby_columns:
            raise ValueError(f"Invalid groupby columns: {invalid_groupby_columns}")
        group_keys = group_keys.distinct()
        self._dataframe = dataframe
        self._group_keys = group_keys
        self._empty = False
        if not group_keys.first():
            if group_keys.columns:
                raise ValueError(
                    "Group keys cannot have no rows, unless it also has no columns"
                )
            self._empty = True
        self._groupby_columns = group_keys.columns

    @property
    def group_keys(self) -> DataFrame:
        """Returns DataFrame containing group keys."""
        return self._group_keys

    @property
    def groupby_columns(self) -> List[str]:
        """Returns DataFrame containing group keys."""
        return self._groupby_columns.copy()

    def select(self, columns: List[str]) -> "GroupedDataFrame":
        """Returns a new GroupedDataFrame object with specified subset of columns.

        Note:
            `columns` must contain the groupby columns.

        Args:
            columns: List of column names to keep. This must include the groupby
                columns.
        """
        if len(set(columns)) != len(columns):
            raise ValueError(f"List contains duplicate column names: {columns}")
        if not set(self.groupby_columns) <= set(columns):
            raise ValueError("Groupby columns must be selected.")
        invalid_columns = [
            column for column in columns if column not in self._dataframe.columns
        ]
        if invalid_columns:
            raise ValueError(f"Invalid columns: {invalid_columns}")
        return GroupedDataFrame(
            dataframe=self._dataframe.select(*columns), group_keys=self.group_keys
        )

    def agg(self, func: Column, fill_value: Any) -> DataFrame:
        """Applies given spark function (column expression) to each group.

        The output DataFrame is guaranteed to have exactly one row for each group
        key. For group keys corresponding to empty groups, the output column will
        contain the supplied `fill_value`. The output DataFrame is also sorted by
        the groupby columns.

        Args:
            func: Function to apply to each group.
            fill_value: Output value for empty groups.
        """
        # pylint: disable=no-member
        if self._empty:
            return self._dataframe.agg(func)
        nonempty_groups_output = self._dataframe.groupBy(self.groupby_columns).agg(func)
        agg_output_columns = set(nonempty_groups_output.columns) - set(
            self.groupby_columns
        )
        assert len(agg_output_columns) == 1
        output_column = agg_output_columns.pop()
        empty_indicator = get_nonconflicting_string(nonempty_groups_output.columns)

        nonempty_groups_output = (
            self._dataframe.groupBy(self.groupby_columns)
            .agg(func)
            .withColumn(empty_indicator, sf.lit(0))
        )
        all_groups_output = join(
            left=self.group_keys,
            right=nonempty_groups_output,
            on=self.groupby_columns,
            how="left",
            nulls_are_equal=True,
        ).fillna({empty_indicator: 1})
        return all_groups_output.withColumn(
            output_column,
            sf.when(sf.col(empty_indicator) == 1, sf.lit(fill_value)).otherwise(
                sf.col(output_column)
            ),
        ).drop(empty_indicator)

    def apply_in_pandas(
        self,
        aggregation_function: Callable[[pd.DataFrame], pd.DataFrame],
        aggregation_output_schema: StructType,
    ) -> DataFrame:
        """Returns DataFrame obtained by applying aggregation function to each group.

        Each group is passed to the `aggregation_function` as a pandas DataFrame
        and the returned pandas DataFrames are stacked into a single spark DataFrame.

        The output DataFrame is guaranteed to have exactly one row for each group
        key. For group keys corresponding to empty groups, the aggregation function
        is applied to an empty pandas DataFrame with the expected schema. The output
        DataFrame is also sorted by the groupby columns.

        Args:
            aggregation_function: Aggregation function to be applied to each group.
            aggregation_output_schema: Expected spark schema for the output of the
                aggregation function.
        """
        spark = SparkSession.builder.getOrCreate()
        if not self.groupby_columns:
            return spark.createDataFrame(
                aggregation_function(self._dataframe.toPandas()),
                schema=aggregation_output_schema,
            )

        empty_indicator = get_nonconflicting_string(self._dataframe.columns)
        sdf = self._dataframe.withColumn(
            empty_indicator, sf.lit(0)  # pylint: disable=no-member
        )

        sdf = join(
            left=self.group_keys,
            right=sdf,
            on=self.groupby_columns,
            how="left",
            nulls_are_equal=True,
        ).fillna({empty_indicator: 1})

        grouped_df = sdf.groupby(*self.groupby_columns)
        agg_input_columns = list(
            set(sdf.columns) - set(self.groupby_columns) - {empty_indicator}
        )
        _wrapper = _create_aggregation_wrapper(
            aggregation_function=aggregation_function,
            empty_indicator=empty_indicator,
            output_schema=aggregation_output_schema,
            groupby_columns=self.groupby_columns,
            input_columns=agg_input_columns,
        )
        output_schema = reduce(
            lambda st, sf: st.add(sf), aggregation_output_schema, self.group_keys.schema
        )

        return grouped_df.applyInPandas(_wrapper, output_schema)

    def get_groups(self) -> Dict[Row, DataFrame]:
        """Returns the groups as dictionary of DataFrames."""
        # pylint: disable=no-member
        groups = {}
        non_grouping_columns = [
            column
            for column in self._dataframe.columns
            if column not in self.groupby_columns
        ]
        for row in self.group_keys.toLocalIterator():
            groups[row] = self._dataframe.filter(
                functools.reduce(
                    lambda acc, x: acc & x,
                    map(
                        lambda x, k=row: (  # type: ignore
                            sf.col(x).eqNullSafe(sf.lit(k[x]))
                        ),
                        row.asDict().keys(),
                    ),
                )
            ).select(non_grouping_columns)
        return groups


def _create_aggregation_wrapper(
    aggregation_function: Callable[[pd.DataFrame], pd.DataFrame],
    empty_indicator: str,
    output_schema: StructType,
    groupby_columns: List[str],
    input_columns: List[str],
):
    """Returns a wrapper function for aggregating all (including empty) groups."""

    def _wrapper(df: pd.DataFrame) -> pd.DataFrame:
        """Remove groupby columns and reorder for the aggregation function."""
        udf_output_columns = output_schema.fieldNames()
        group_value = df.loc[0, groupby_columns].to_dict()

        # If the group didn't exist in the original dataframe, pass an empty
        # dataframe to the aggregation function.
        if df.loc[0, empty_indicator]:
            df = pd.DataFrame(columns=input_columns)
        else:
            df = df.loc[:, input_columns]

        aggregated_df = aggregation_function(df)

        # Add back groupby columns.
        for groupby_column, value in group_value.items():
            aggregated_df[groupby_column] = value

        # Fix column ordering when returning.
        return aggregated_df.loc[:, list(group_value) + udf_output_columns]

    return _wrapper



"""Measurements constructed by chaining other measurements and transformations."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from typing import Any, Callable, Optional

from typeguard import typechecked

from tmlt.core.exceptions import DomainMismatchError, MetricMismatchError
from tmlt.core.measurements.base import Measurement
from tmlt.core.transformations.base import Transformation


class ChainTM(Measurement):
    """Measurement constructed by chaining a transformation and a measurement."""

    @typechecked
    def __init__(
        self,
        transformation: Transformation,
        measurement: Measurement,
        hint: Optional[Callable[[Any, Any], Any]] = None,
    ):
        """Constructor.

        Args:
            transformation: Transformation component (called first).
            measurement: Measurement component (called second).
            hint: An optional function to compute the intermediate metric value
                (after the transformation, but before the measurement) for
                :meth:`~.privacy_relation`. It takes in the same inputs as
                :meth:`~.privacy_relation`, and is only required if the
                transformation's :meth:`~.Transformation.stability_function` raises
                NotImplementedError.
        """
        if transformation.output_domain != measurement.input_domain:
            raise DomainMismatchError(
                (transformation.output_domain, measurement.input_domain),
                (
                    "Can not chain transformation and measurement: Mismatching"
                    f" domains.\n{transformation.output_domain} !="
                    f" {measurement.input_domain}"
                ),
            )

        if transformation.output_metric != measurement.input_metric:
            raise MetricMismatchError(
                (transformation.output_metric, measurement.input_metric),
                (
                    "Can not chain transformation and measurement: Mismatching"
                    f" metrics.\n{transformation.output_metric} !="
                    f" {measurement.input_metric}"
                ),
            )
        super().__init__(
            input_domain=transformation.input_domain,
            input_metric=transformation.input_metric,
            output_measure=measurement.output_measure,
            is_interactive=measurement.is_interactive,
        )
        self._measurement = measurement
        self._transformation = transformation
        self._hint = hint

    @property
    def measurement(self) -> Measurement:
        """Returns measurement being chained."""
        return self._measurement

    @property
    def transformation(self) -> Transformation:
        """Returns transformation being chained."""
        return self._transformation

    @typechecked
    def privacy_function(self, d_in: Any) -> Any:
        """Returns the smallest d_out satisfied by the measurement.

        Returns M.privacy_function(T.stability_function(d_in)).

        where:

        * T is the transformation applied (:attr:`~ChainTM.transformation`")
        * M is the measurement applied (:attr:`~ChainTM.measurement`")

        Args:
            d_in: Distance between inputs under input_metric.

        Raises:
            NotImplementedError: If M.privacy_function(T.stability_function(d_in))
                raises :class:`NotImplementedError`.
        """
        return self.measurement.privacy_function(
            self.transformation.stability_function(d_in)
        )

    @typechecked
    def privacy_relation(self, d_in: Any, d_out: Any) -> bool:
        """Returns True only if outputs are close under close inputs.

        Let d_mid = T.stability_function(d_in), or hint(d_in, d_out) if
        T.stability_function raises :class:`NotImplementedError`.

        This returns True only if the following hold:

        (1) T.stability_relation(d_in, d_mid)
        (2) M.privacy_relation(d_mid, d_out)

        where:

        * T is the transformation applied (:attr:`~ChainTM.transformation`")
        * M is the measurement applied (:attr:`~ChainTM.measurement`")
        * hint is the hint passed to the constructor.

        Args:
            d_in: Distance between inputs under input_metric.
            d_out: Distance between outputs under output_measure.

        Raises:
            ValueError: If a hint is not provided and T.stability_function raises
                :class:`NotImplementedError`.
        """
        self.input_metric.validate(d_in)
        self.output_measure.validate(d_out)
        try:
            d_mid = self.transformation.stability_function(d_in)
        except NotImplementedError as e:
            if self._hint is None:
                raise ValueError(
                    "A hint is needed to check this privacy relation, because the "
                    "stability_relation of self.transformation raised a "
                    f"NotImplementedError: {e}"
                ) from e
            d_mid = self._hint(d_in, d_out)
        return self.transformation.stability_relation(
            d_in, d_mid
        ) and self.measurement.privacy_relation(d_mid, d_out)

    def __call__(self, data: Any) -> Any:
        """Computes measurement after applying transformation on input data."""
        return self._measurement(self._transformation(data))
